#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import base64
import getpass
import os
import time
from collections import defaultdict

from ._version import __version__
from .controller import request_answer_from_controller
from .utils import dot_to_gif, env, format_duration


class WorkflowSig(object):

    def __init__(self, workflow_id):
        self.data = defaultdict(lambda: defaultdict(list))

        for entry_type, id, item in request_answer_from_controller(
            ["workflow_sig", "records", workflow_id]):
            try:
                self.data[entry_type][id].append(item.strip())
            except Exception as e:
                env.logger.debug(f"Failed to read report line: {e}")

    def convert_time(self, info):
        for key in [x for x in info.keys() if x.endswith("_time")]:
            info[key + "_str"] = time.strftime("%Y-%m-%d %H:%M:%S",
                                               time.localtime(info[key]))
        if "start_time" in info and "end_time" in info:
            info["duration"] = int(info["end_time"] - info["start_time"])
            info["duration_str"] = format_duration(info["duration"])
        return info

    def workflows(self):
        try:
            # workflows has format
            # workflow  workflow_id dict1
            # workflow  workflow_id dict2
            # workflow  workflow_id2 dict1
            workflows = defaultdict(dict)
            for id, values in self.data["workflow"].items():
                for val in values:
                    workflows[id].update(eval(val))
            for v in workflows.values():
                self.convert_time(v)
                if "dag" in v:
                    try:
                        v["dag_img"] = dot_to_gif(
                            v["dag"], warn=env.logger.warning)
                    except Exception as e:
                        env.logger.warning(
                            f"Failed to obtain convert dag to image: {e}")
                if "script" in v:
                    v["script"] = base64.b64decode(v["script"]).decode()
            return workflows
        except Exception as e:
            env.logger.warning(f"Failed to obtain workflow information: {e}")
            return {}

    def tasks(self):

        def merge_dict(items):
            ret = {}
            for item in items:
                ret.update(eval(item))
            return ret

        try:
            # there can be multiple task status for each id
            tasks = {
                id: merge_dict(res) for id, res in self.data["task"].items()
            }
            for val in tasks.values():
                self.convert_time(val)
                if "peak_cpu" in val:
                    val["peak_cpu_str"] = f'{val["peak_cpu"]:.1f}%'
                if "peak_mem" in val:
                    val["peak_mem_str"] = f'{val["peak_mem"] / 1024 / 1024 :.1f}Mb'
            return tasks
        except Exception:
            return {}

    def steps(self):
        try:
            return {
                wf: [self.convert_time(eval(x)) for x in steps
                    ] for wf, steps in self.data["step"].items()
            }
        except Exception as e:
            env.logger.warning(str(e))
            return {}

    def transcripts(self):
        try:
            return {
                step: [self.convert_time(eval(x)) for x in items
                      ] for step, items in self.data["transcript"].items()
            }
        except Exception as e:
            env.logger.warning(str(e))
            return {}

    def placeholders(self):
        try:
            return self.data["placeholder"]["file_target"]
        except Exception:
            return []


def calc_timeline(info, start_time, total_duration):
    if total_duration == 0 or "start_time" not in info or "end_time" not in info:
        info["before_percent"] = 0
        info["during_percent"] = 100
        info["after_precent"] = 0
        return
    info["before_percent"] = int(
        (info["start_time"] - start_time) * 100 / total_duration)
    info["during_percent"] = max(1,
                                 int(info["duration"] * 100 / total_duration))
    info[
        "after_percent"] = 100 - info["before_percent"] - info["during_percent"]
    return info


def render_report(output_file, workflow_id):
    data = WorkflowSig(workflow_id)

    from jinja2 import Environment, PackageLoader, select_autoescape

    environment = Environment(
        loader=PackageLoader("sos", "templates"),
        autoescape=select_autoescape(["html", "xml"]),
    )
    environment.filters["basename"] = os.path.basename
    template = environment.get_template("workflow_report.tpl")

    context = {
        "workflows": data.workflows(),
        "tasks": data.tasks(),
        "steps": data.steps(),
        "transcripts": data.transcripts(),
        "sos_version": __version__,
        "user": getpass.getuser(),
        "time_now_str": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
    }
    # derived context
    context["master_id"] = next(iter(
        context["workflows"].values()))["master_id"]
    try:
        # calculate percentage
        start_time = context["workflows"][context["master_id"]]["start_time"]
        total_duration = (
            context["workflows"][context["master_id"]]["end_time"] - start_time)
        for info in context["workflows"].values():
            calc_timeline(info, start_time, total_duration)
        for steps in context["steps"].values():
            for step in steps:
                calc_timeline(step, start_time, total_duration)
        for info in context["tasks"].values():
            calc_timeline(info, start_time, total_duration)
        with open(output_file, "w") as wo:
            wo.write(template.render(context))
        env.logger.info(f"Summary of workflow saved to {output_file}")
    except Exception as e:
        env.logger.error(f"Failed to generate report {output_file}: {e}")



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import os
import shutil
import subprocess
import unittest

from sos import execute_workflow
from sos.parser import SoS_Script
from sos.targets import file_target
from sos.utils import env
from sos.workflow_executor import Base_Executor


class TestNested(unittest.TestCase):

    def setUp(self):
        env.reset()
        subprocess.call('sos remove -s', shell=True)
        self.temp_files = []

    def tearDown(self):
        for f in self.temp_files:
            if file_target(f).exists():
                file_target(f).unlink()

    def touch(self, files):
        '''create temporary files'''
        if isinstance(files, str):
            files = [files]
        #
        for f in files:
            with open(f, 'w') as tmp:
                tmp.write('test')
        #
        self.temp_files.extend(files)

    def test_progress_bar(self):
        # progress bar with nested workflow
        script = SoS_Script('''
import time
time.sleep(0)
[sub_1]
[sub_2]
[sub_3]
[sub_4]
[a_1]
[a_2]
[a_3]
sos_run('sub')
[a_4]
[a_5]
''')
        env.verbosity = 1
        wf = script.workflow('a')
        Base_Executor(wf).run()

    def test_nested_workflow(self):
        '''Test the creation and execution of combined workfow'''
        self.touch(['a.txt', 'b.txt', 'b.begin'])
        script = SoS_Script('''
if 'executed' not in locals():
    executed = []
if 'inputs' not in locals():
    inputs = []

[a_1: shared=['executed', 'inputs']]
executed.append(step_name)
inputs.append(_input)

[a_2: shared=['executed', 'inputs']]
executed.append(step_name)
inputs.append(_input)

[a_3: shared=['executed', 'inputs']]
executed.append(step_name)
inputs.append(_input)

[a_4: shared=['executed', 'inputs']]
executed.append(step_name)
output: 'a.done'
inputs.append(_input)
run: expand=True
  touch {_output}

[b_1: shared=['executed', 'inputs']]
executed.append(step_name)
input: 'b.begin'
inputs.append(_input)

[b_2: shared=['executed', 'inputs']]
executed.append(step_name)
inputs.append(_input)

[b_3: shared=['executed', 'inputs']]
executed.append(step_name)
inputs.append(_input)

[b_4: shared=['executed', 'inputs']]
executed.append(step_name)
output: 'b.txt'
inputs.append(_input)

[c: shared=['executed', 'inputs']]
executed.append(step_name)
input: 'a.txt'
output: 'b.txt'
inputs.append(_input)
sos_run('a+b', shared=['executed', 'inputs'])
''')
        env.config['sig_mode'] = 'ignore'
        wf = script.workflow('c')
        Base_Executor(wf).run()
        # order of execution is not guaranteed
        self.assertEqual(
            sorted(env.sos_dict['executed']),
            sorted(
                ['c', 'a_1', 'a_2', 'a_3', 'a_4', 'b_1', 'b_2', 'b_3', 'b_4']))
        env.sos_dict.pop('executed', None)

    def test_looped_nested_workflow(self):
        # step will be looped
        self.touch(['a.txt', 'b.txt'])
        script = SoS_Script('''
if 'executed' not in locals():
    executed = []
if 'inputs' not in locals():
    inputs = []

[a_1:shared=['executed', 'inputs']]
executed.append(step_name)
output: _input[0] + '.a1'
inputs.append(_input)
run: expand=True
    touch {_output}

[a_2:shared=['executed', 'inputs']]
executed.append(step_name)
output: _input[0] + '.a2'
inputs.append(_input)
run: expand=True
    touch {_output}

[c:shared=['executed', 'inputs']]
executed.append(step_name)
input: 'a.txt', 'b.txt', group_by='single'
inputs.append(_input)
sos_run('a', shared=['executed', 'inputs'])
''')
        wf = script.workflow('c')
        Base_Executor(wf).run()
        self.assertEqual(env.sos_dict['executed'],
                         ['c', 'a_1', 'a_2', 'a_1', 'a_2'])
        #self.assertEqual(env.sos_dict['inputs'], [['a.txt'], ['a.txt'], ['a.txt.a1'], ['b.txt'], ['b.txt'], ['b.txt.a1']])
        for file in ('a.txt.a1', 'a.txt.a1.a2', 'b.txt.a1', 'b.txt.a1.a2'):
            if file_target(file).exists():
                file_target(file).unlink()
        #

    def test_single_looped_nested_workflow(self):
        self.touch(['a.txt', 'b.txt'])
        env.sos_dict.pop('executed', None)
        # allow specifying a single step
        # step will be looped
        script = SoS_Script('''
if 'executed' not in locals():
    executed = []
[a_1:shared='executed']
executed.append(step_name)
[a_2:shared='executed']
executed.append(step_name)
[c_0:shared='executed']
executed.append(step_name)
[c_1:shared='executed']
depends: sos_variable('executed')
executed.append(step_name)
input: 'a.txt', 'b.txt', group_by='single'
sos_run('a:2', shared='executed')
''')
        wf = script.workflow('c')
        Base_Executor(wf).run()
        self.assertEqual(env.sos_dict['executed'], ['c_0', 'c_1', 'a_2', 'a_2'])
        env.sos_dict.pop('executed', None)
        # allow specifying a single step
        # step will be looped
        script = SoS_Script('''
if 'executed' not in locals():
    executed = []
[a_1:shared='executed']
executed.append(step_name)
[a_2:shared='executed']
executed.append(step_name)
[c_0:shared='executed']
executed.append(step_name)
[c_1:shared='executed']
depends: sos_variable('executed')
executed.append(step_name)
input: 'a.txt', 'b.txt', group_by='single'
sos_run('a:2', shared='executed')
''')
        wf = script.workflow('c')
        Base_Executor(wf).run()
        self.assertEqual(env.sos_dict['executed'], ['c_0', 'c_1', 'a_2', 'a_2'])
        #
        env.sos_dict.pop('executed', None)

    def test_recursive_nested_workflow(self):
        # recursive subworkflow not allowed
        self.touch(['a.txt', 'b.txt'])
        script = SoS_Script('''
if 'executed' not in locals():
    executed = []
[a_1:shared='executed']
executed.append(step_name)
[a_2:shared='executed']
executed.append(step_name)
[c_0:shared='executed']
executed.append(step_name)
[c_1:shared='executed']
executed.append(step_name)
input: 'a.txt', 'b.txt', group_by='single'
sos_run('a_2+c', shared='executed')
''')
        wf = script.workflow('c')
        self.assertRaises(Exception, Base_Executor(wf).run)
        #
        env.sos_dict.pop('executed', None)
        # nested subworkflow is allowed
        script = SoS_Script('''
if 'executed' not in locals():
    executed = []
[a_1:shared='executed']
executed.append(step_name)
[a_2:shared='executed']
executed.append(step_name)
[a_3:shared='executed']
executed.append(step_name)
[b_1:shared='executed']
executed.append(step_name)
[b_2:shared='executed']
executed.append(step_name)
sos_run('a:1-2', shared='executed')
[c_0:shared='executed']
executed.append(step_name)
[c_1:shared='executed']
depends: sos_variable('executed')
executed.append(step_name)
input: 'a.txt'
sos_run('a+b', shared='executed')
''')
        wf = script.workflow('c')
        Base_Executor(wf).run()
        self.assertEqual(
            env.sos_dict['executed'],
            ['c_0', 'c_1', 'a_1', 'a_2', 'a_3', 'b_1', 'b_2', 'a_1', 'a_2'])
        #
        #
        env.sos_dict.pop('executed', None)

    def test_subworkflow_with_options(self):
        # nested subworkflow with step option and others
        self.touch(['a.txt', 'b.txt'])
        script = SoS_Script('''
if 'executed' not in locals():
    executed = []
[a_1:shared='executed']
executed.append(step_name)
[a_2:shared='executed']
executed.append(step_name)
[a_3:shared='executed']
executed.append(step_name)
[b:shared='executed']
executed.append(step_name)
input: 'a.txt', 'b.txt', group_by='single'
sos_run('a:3+a:1', shared='executed')
[d:shared='executed']
executed.append(step_name)
input: 'a.txt', 'b.txt', group_by='single'
sos_run('a:2', shared='executed')
[e2_2:shared='executed']
executed.append(step_name)
input: 'a.txt', 'b.txt', group_by='single'
''')
        wf = script.workflow('b')
        Base_Executor(wf).run()
        self.assertEqual(env.sos_dict['executed'],
                         ['b', 'a_3', 'a_1', 'a_3', 'a_1'])
        env.sos_dict.pop('executed', None)
        wf = script.workflow('d')
        Base_Executor(wf).run()
        self.assertEqual(env.sos_dict['executed'], ['d', 'a_2', 'a_2'])
        env.sos_dict.pop('executed', None)
        wf = script.workflow('e2')
        Base_Executor(wf).run()
        self.assertEqual(env.sos_dict['executed'], ['e2_2'])
        #
        # clean up
        file_target('a.done').unlink()

    def test_dynamic_nested_workflow(self):
        '''Test nested workflow controlled by command line option'''
        script = SoS_Script('''
if 'executed' not in locals():
    executed = []
parameter: wf='a'

[a_1:shared='executed']
executed.append(step_name)
[a_2:shared='executed']
executed.append(step_name)
[a_3:shared='executed']
executed.append(step_name)
[b_1:shared='executed']
executed.append(step_name)
[b_2:shared='executed']
executed.append(step_name)
[b_3:shared='executed']
executed.append(step_name)

[default:shared='executed']
executed.append(step_name)
sos_run(wf, shared='executed')
''')
        wf = script.workflow()
        Base_Executor(wf, args=['--wf', 'b']).run()
        self.assertEqual(env.sos_dict['executed'],
                         ['default', 'b_1', 'b_2', 'b_3'])
        #
        env.sos_dict.pop('executed', None)
        Base_Executor(wf, args=['--wf', 'a']).run()
        self.assertEqual(env.sos_dict['executed'],
                         ['default', 'a_1', 'a_2', 'a_3'])

    def test_sos_run(self):
        '''Test action sos_run with keyword parameters'''
        for f in ['0.txt', '1.txt']:
            if file_target(f).exists():
                file_target(f).unlink()
        script = SoS_Script(r'''
[A]
parameter: num=5
run: expand=True
    touch {num}.txt

[batch]
for k in range(2):
    sos_run('A', num=k)
''')
        wf = script.workflow('batch')
        Base_Executor(wf).run()
        for f in ['0.txt', '1.txt']:
            self.assertTrue(file_target(f).target_exists())
            file_target(f).unlink()
        #
        # if we do not pass num, parameter would not change
        for f in ['0.txt', '1.txt']:
            if file_target(f).exists():
                file_target(f).unlink()
        script = SoS_Script(r'''
[A]
parameter: num=5
run: expand=True
    touch {num}.txt

[batch]
for num in range(2):
    sos_run('A')
''')
        wf = script.workflow('batch')
        Base_Executor(wf).run()
        for f in ['0.txt', '1.txt']:
            self.assertFalse(file_target(f).target_exists())
        self.assertTrue(file_target('5.txt').target_exists())
        file_target('5.txt').unlink()
        #
        # test parameter shared to send and return vars
        #
        script = SoS_Script(r'''
[A: shared='k']
k += 10

[batch]
for k in range(2):
    sos_run('A', shared='k')
    run(f"touch {k}.txt")
''')
        wf = script.workflow('batch')
        Base_Executor(wf).run()
        for f in ['10.txt', '11.txt']:
            self.assertTrue(file_target(f).target_exists())
            file_target(f).unlink()

    def test_da_gof_dynamic_nested_workflow(self):
        #
        # Because we are not sure which workflows would be executed
        # until run time, the DAG should not contain nested workflow
        # until runtime.
        #
        for f in [
                'B0.txt', 'B0.txt.p', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p'
        ]:
            if file_target(f).exists():
                file_target(f).unlink()
        #
        #  A1 <- P <- B
        #  A1 <- P <- B
        #  A2
        #
        #  ALL calls A and B with parameter
        #
        script = SoS_Script('''
[A_1]
parameter: num = 2
input: f"B{num}.txt.p"

[B: provides='B{num}.txt']
run: expand=True
    touch 'B{num[0]}.txt'

[P: provides='{filename}.p']
input: filename
run: expand=True
    touch {_output}

[ALL]

for i in range(3):
    sos_run('A', num=i)


''')
        # the workflow should call step K for step C_2, but not C_3
        wf = script.workflow('ALL')
        Base_Executor(wf).run()
        for f in [
                'B0.txt', 'B0.txt.p', 'B1.txt', 'B1.txt.p', 'B2.txt', 'B2.txt.p'
        ]:
            self.assertTrue(file_target(f).target_exists())
            file_target(f).unlink()

    def test_outcome_oriented_nested_workflow(self):
        '''test nested workflow triggered by targets'''
        if os.path.isfile('test_15.txt'):
            os.remove('test_15.txt')
        execute_workflow('''
[A: provides='test_{idx}.txt']
_output.touch()

[default]
sos_run(targets='test_15.txt')
        ''')
        self.assertTrue(os.path.isfile('test_15.txt'))

    def test_passing_vars_to_nested_workflow(self):
        '''Test if variables can be passed to nested workflows'''
        script = SoS_Script(r"""
import time
import random

[nested]
parameter: nested=True
parameter: seed=1
print(f'I am nested {nested} with seed {seed}')

[0]
reps = range(5)
input: for_each='reps'
import random
nested = _reps
seed = random.randint(1, 1000)
print(f'Passing {seed} to {nested}')
sos_run('nested', nested=nested, seed=seed)

""")
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_user_defined_func(self):
        '''Test the use of user-defined functions in SoS script'''
        script = SoS_Script(r"""

def myfunc():
  return 'a'

[1: shared={'test':'_output'}]
output: myfunc()

myfunc()

""")
        wf = script.workflow()
        Base_Executor(wf).run(mode='dryrun')
        self.assertEqual(env.sos_dict['test'], ['a'])
        # User defined function should also work under nested workflows
        # This is difficult because the 'local namespace' is usually
        # not seen inside function definition. The solution now is to
        # use a single workspace.
        script = SoS_Script(r"""

def myfunc():
    # test if builtin functions (sum and range) can be used here.
    return 'a' + str(sum(range(10)))

[1: shared={'test':'_output'}]
output: [myfunc() for i in range(10)][0]

myfunc()

""")
        wf = script.workflow()
        Base_Executor(wf).run(mode='dryrun')
        self.assertEqual(env.sos_dict['test'], ['a45'])

    def test_config_file_of_nested_workflow(self):
        '''Test passing of configurationg to nested workflow'''
        script = SoS_Script('''
[test_1]
parameter: key = None
print(CONFIG[key])

[default_1]
sos_run('test:1', key = '1')
    ''')
        with open('test.conf', 'w') as conf:
            conf.write("""{'1':'hi'}""")
        wf = script.workflow()
        Base_Executor(wf, config={'config_file': 'test.conf'}).run()

    def test_error_from_subworkflow(self):
        '''Test if error from subworkflow is passed to master (#396)'''
        script = SoS_Script('''
[test_1]
R:
  set.seed(xxx)

[default]
sos_run('test')
''')
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)

    def test_fun_def(self):
        '''Test defintion of function that can be used by other steps'''
        self.touch(['aa.txt', 'ab.txt'])
        # in nested workflow?
        script = SoS_Script(r"""
def myfunc(a):
    return ['a' + x for x in a]

[mse: shared={'test':'_output'}]
input: myfunc(['a.txt', 'b.txt'])

[1]
sos_run('mse')
""")
        wf = script.workflow()
        Base_Executor(wf).run(mode='dryrun')
        #
        # Names defined in subworkflow is not returned to the master dict
        self.assertTrue('test' not in env.sos_dict)

    def test_search_path(self):
        '''Test if any action should exit in five seconds in dryrun mode'''
        sos_config_file = os.path.join(
            os.path.expanduser('~'), '.sos', 'config.yml')
        shutil.copy(sos_config_file, 'test.yml')
        #
        subprocess.call(
            'sos config --set sos_path {0}/crazy_path {0}/crazy_path/more_crazy/'
            .format(os.getcwd()),
            shell=True)
        #
        if not os.path.isdir('crazy_path'):
            os.mkdir('crazy_path')
            os.mkdir(os.path.join('crazy_path', 'more_crazy'))
        with open(os.path.join('crazy_path', 'crazy_master.sos'), 'w') as crazy:
            crazy.write('''
[0]
sos_run('cc', source='crazy_slave.sos')

''')
        with open(
                os.path.join('crazy_path', 'more_crazy', 'crazy_slave.sos'),
                'w') as crazy:
            crazy.write('''
[cc_0]
print('hay, I am crazy')
''')

        script = SoS_Script(filename='crazy_master.sos')
        script.workflow()
        #
        shutil.rmtree('crazy_path')
        os.remove(sos_config_file)
        shutil.copy('test.yml', sos_config_file)

    def test_nested_workdir(self):
        '''Test nested runtime option for work directory'''
        if os.path.isdir('tmp'):
            shutil.rmtree('tmp')
        script = SoS_Script('''
[step]
task: workdir='tmp'
run:
    touch 'a.txt'

[default]
sos_run('step', workdir='tmp')
''')
        wf = script.workflow()
        # this should be ok.
        Base_Executor(wf, config={'default_queue': 'localhost'}).run()
        os.path.isfile('tmp/tmp/a.txt')
        shutil.rmtree('tmp')

    def test_failure_of_nested_workflow(self):
        '''Test failure of nested workflow #838'''
        if os.path.isfile('a.txt'):
            os.remove('a.txt')
        script = SoS_Script('''
[something]
input: 'a.txt'

[default]
sos_run('something')
''')
        wf = script.workflow()
        # this should be ok.
        self.assertRaises(Exception, Base_Executor(wf).run)

    def test_nested_from_another_file(self):
        '''Test nested runtime option for work directory'''
        if os.path.isdir('a.txt'):
            os.remove('a.txt')
        with open('another.sos', 'w') as another:
            another.write('''
[whatever]
run:
    touch 'a.txt'

''')
        script = SoS_Script('''
[default]
sos_run('whatever', source='another.sos')
''')
        wf = script.workflow()
        # this should be ok.
        Base_Executor(wf).run()
        self.assertTrue(
            os.path.isfile('a.txt'),
            'a.txt should have been created by nested workflow from another file'
        )

    def test_concurrent_sub_workflow(self):
        '''Test concurrent subworkflow sos_run '''
        script = SoS_Script('''
[A]
parameter: idx=0
import time
time.sleep(5)

[default]
input: for_each=dict(i=range(6))
sos_run('A', idx=i)
''')
        import time
        st = time.time()
        wf = script.workflow()
        # this should be ok.
        Base_Executor(wf, config={'worker_procs': ['8']}).run()
        self.assertTrue(time.time() - st < 30)

    def test_sos_multi_workflow(self):
        '''Test multiple workflows in sos_run '''
        script = SoS_Script('''
[B]
parameter: idx=2
import time
time.sleep(idx)

[A]
import time
time.sleep(idx)

[default]
input: for_each=dict(i=range(4))
sos_run(['A', 'B'], idx=i)
''')
        import time
        st = time.time()
        wf = script.workflow()
        # this should be ok.
        Base_Executor(wf, config={'worker_procs': ['8']}).run()
        self.assertTrue(time.time() - st < 20)

    def test_pass_of_args(self):
        '''Test passing of arguments through sos_run #1164'''
        script = SoS_Script('''
parameter: b=2
[A]
parameter: c=1
print(c)

[default]
sos_run('A', a=2)
''')
        wf = script.workflow()
        self.assertRaises(Exception, Base_Executor(wf).run)
        #
        script = SoS_Script('''
parameter: b=2
[A]
parameter: c=1
print(c)

[default]
sos_run('A', b=2)
''')
        wf = script.workflow()
        Base_Executor(wf).run()
        #
        script = SoS_Script('''
parameter: b=2
[A]
parameter: c=1
print(c)

[default]
sos_run('A', c=2)
''')
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_nested_dynamic_depends(self):
        '''Test the execution of nested workflow with dynamic depends'''
        script = SoS_Script('''
[A_1]
parameter: num = 20
input: f"B{num}.txt"
depends: _input + '.p'

[B: provides='B{num}.txt']
run: expand=True
    touch {_output}

[P: provides='{filename}.p']
input: filename
run: expand=True
    touch {_output}


[default]
sos_run('A', num=0)
''')
        wf = script.workflow()
        Base_Executor(wf).run()

    def test_neste_with_both_workflow_and_targets(self):
        '''Test nested workflow with both workflow and targets'''
        for file in ('A_out.txt', 'B_out.txt'):
            if os.path.isfile(file):
                os.remove(file)
        execute_workflow(r'''
[A_20]
output: 'A_out.txt'
_output.touch()

[B]
output: 'B_out.txt'
_output.touch()

[default]
sos_run('A', targets='B_out.txt')
''')
        self.assertTrue(os.path.isfile('A_out.txt'))
        self.assertTrue(os.path.isfile('B_out.txt'))


if __name__ == '__main__':
    #suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestParser)
    # unittest.TextTestRunner(, suite).run()
    unittest.main()



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import subprocess
import sys

import pexpect

from .hosts import Host
from .targets import path
from .utils import env


def list_queues(cfg, hosts=[], verbosity=1):
    env.verbosity = 1
    all_hosts = cfg.get("hosts", [])
    if not all_hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    for host in hosts:
        if host not in all_hosts:
            env.logger.warning(f"Undefined host {host}")
    host_description = [
        ["Alias", "Address", "Queue Type", "Description"],
        ["-----", "-------", "----------", "-----------"],
    ]
    for host in sorted([x for x in hosts
                        if x in all_hosts] if hosts else all_hosts):
        try:
            h = Host(host, start_engine=False, test_connection=False)
        except Exception as e:
            if verbosity == 0:
                print(f"{host} ({e})")
            elif verbosity in (1, 2):
                host_description.append([host, "?", "?", str(e)])
            else:
                print(f"Queue:       {host}")
                print(f"Error:       {str(e)}")
                if isinstance(cfg["hosts"][host], dict):
                    print("Configuration:")
                    for key in cfg["hosts"][host].keys():
                        print(
                            f'  {(key + ":").ljust(24)} {cfg["hosts"][host][key]}'
                        )
                print()
            continue
        if verbosity == 0:
            print(h.alias)
        elif verbosity in (1, 2):
            host_description.append(
                [h.alias, h._host_agent.address, h._engine_type, h.description])
        else:
            print(f"Queue:       {h.alias}")
            print(f"Address:     {h._host_agent.address}")
            print(f"Queue Type:  {h._engine_type}")
            print(f"Description: {h.description}")
            print("Configuration:")
            keys = sorted(h.config.keys())
            for key in keys:
                print(f'  {(key + ":").ljust(24)} {h.config[key]}')
            print()
    if verbosity in (1, 2):
        width = [(len(x) for x in row) for row in host_description]
        max_width = [max(x) for x in zip(*width)]
        print("\n".join(" ".join([t.ljust(w)
                                  for t, w in zip(row, max_width)])
                        for row in host_description))


def status_of_queues(cfg, hosts=[], verbosity=1):
    env.verbosity = 1
    all_hosts = cfg.get("hosts", [])
    if not all_hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    for host in hosts:
        if host not in all_hosts:
            env.logger.warning(f"Undefined host {host}")
    host_description = [
        ["Alias", "Address", "Queue Type", "Running", "Pending", "Completed"],
        ["-----", "-------", "----------", "-------", "-------", "---------"],
    ]
    for host in sorted([x for x in hosts
                        if x in all_hosts] if hosts else all_hosts):
        try:
            h = Host(host, start_engine=True)
            status = h._task_engine.query_tasks(tasks=[], verbosity=0)
            if not status:
                raise ValueError("Failed to query status")
        except Exception as e:
            if verbosity == 0:
                print(f"{host} ({e})")
            elif verbosity in (1, 2):
                host_description.append([host, "?", "?", "?", "?", "?"])
            else:
                print(f"Queue:       {host}")
                print(f"Error:       {str(e)}")
                if isinstance(cfg["hosts"][host], dict):
                    print("Configuration:")
                    for key in cfg["hosts"][host].keys():
                        print(
                            f'  {(key + ":").ljust(24)} {cfg["hosts"][host][key]}'
                        )
                print()
            continue
        status = [x.strip() for x in status.splitlines() if x.strip()]
        running = str(status.count("running"))
        pending = str(status.count("pending"))
        completed = str(status.count("completed"))

        if verbosity == 0:
            print(f"{h.alias} {running} {pending} {completed}")
        elif verbosity in (1, 2):
            host_description.append([
                h.alias,
                h._host_agent.address,
                h._engine_type,
                running,
                pending,
                completed,
            ])
        else:
            print(f"Queue:       {h.alias}")
            print(f"Address:     {h._host_agent.address}")
            print(f"Queue Type:  {h._engine_type}")
            print(f"Description: {h.description}")
            for k in set(status):
                if k not in ("running", "pending", "completed"):
                    print(f'{(k+":").ljust(26)} {status.count(k)}')
            print("Configuration:")
            keys = sorted(h.config.keys())
            for key in keys:
                print(f'  {(key + ":").ljust(24)} {h.config[key]}')
            print()
    if verbosity in (1, 2):
        width = [(len(x) for x in row) for row in host_description]
        max_width = [max(x) for x in zip(*width)]
        print("\n".join(" ".join([t.ljust(w)
                                  for t, w in zip(row, max_width)])
                        for row in host_description))


def test_ssh(host):
    return host.test_connection()


def test_scp(host):
    if host.address == "localhost":
        return "OK"
    # test send task file
    import random

    tID = random.randint(1, 100000)
    task_filename = os.path.join(
        os.path.expanduser("~"), ".sos", f"test_{tID}.tmp")
    with open(task_filename, "w") as test_task:
        test_task.write("test task")
    # test scp
    try:
        host.send_job_file(task_filename)
    except Exception as e:
        return str(e)
    # test remove file using ssh
    try:
        host.check_call(f"rm -f ~/.sos/test_{tID}.tmp")
    except Exception as e:
        return str(e)
    # test rsync
    return "OK"


def test_cmd(host, cmd):
    # test the execution of sos commands
    try:
        ret = host.check_call(
            cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if ret == 0:
            return "OK"
        else:
            return "sos not installed or not accessible on host."
    except Exception as e:
        return str(e)


def test_paths(host):
    if host.address == "localhost":
        return "OK"
    # shared means, if localhost creates a file, it should be
    # instantly available on the remote host
    if not host.path_map:
        return "No path_map between local and remote host."
    import random

    tID = random.randint(1, 100000)
    for local, remote in host.path_map.items():
        if local in host.shared_dirs:
            # will be tested by 'shared'
            continue
        # now, let us see if two directory has the same files?
        if not os.path.isdir(local):
            return f"Mapped directory {local} does not exist."
        # remote?
        try:
            host.check_output(f"ls -a {path(remote):q}")
        except Exception:
            return f"Failed to access shared directory {remote} on remote host."

        # test if local directory is writable
        try:
            with open(os.path.join(local, f".sos_test_{tID}.txt"),
                      "w") as tFile:
                tFile.write(f"{tID}")
        except Exception:
            return f"Failed to write to mapped directory {local}"
        # test if file can be sent
        try:
            host.send_to_host(os.path.join(local, f".sos_test_{tID}.txt"))
        except Exception as e:
            return (
                f"Failed to send files under {local} to remote host under {remote}: {e}"
            )

        # the file should be available on remote host
        try:
            remote_content = host.check_output(
                f"cat {remote}/.sos_test_{tID}.txt")
        except Exception as e:
            return (
                f"Failed to send files under {local} to remote host under {remote}: {e}"
            )
        if remote_content != str(tID):
            return f"Content of file sent does not match: {tID} sent, {remote_content} received"
        # test retrieving files
        # remove local file
        os.remove(os.path.join(local, f".sos_test_{tID}.txt"))
        # copy file back
        try:
            host.receive_from_host(os.path.join(local, f".sos_test_{tID}.txt"))
        except Exception as e:
            return f"Failed to receive file from remote host {remote}: {e}"
        #
        if not os.path.isfile(os.path.join(local, f".sos_test_{tID}.txt")):
            return (
                f"Failed to receive file from remote host {remote}: file does not exist"
            )
        # check file content?
        with open(os.path.join(local, f".sos_test_{tID}.txt"), "r") as tFile:
            remote_content = tFile.read()
        if remote_content != str(tID):
            return f"Content of received file does not match: {tID} expected, {remote_content} received."
        # if everything ok, remove local and remote test files
        os.remove(os.path.join(local, f".sos_test_{tID}.txt"))
        #
        try:
            remote_content = host.check_output(
                f"rm {remote}/.sos_test_{tID}.txt")
        except Exception as e:
            return f"Failed to remove test file on remote host: {e}"
    return "OK"


def test_shared(host):
    if host.address == "localhost":
        return "OK (localhost)"
    # shared means, if localhost creates a file, it should be
    # instantly available on the remote host
    for local in host.shared_dirs:
        if local not in host.path_map:
            return f"shared directory {local} not in path_map"
        # now, let us see if two directory has the same files?
        if not os.path.isdir(local):
            return f"shared directory {local} does not exist."
        local_files = os.listdir(local)
        # remote?
        remote = host.path_map[local]
        try:
            remote_files = host.check_output(f"ls -a {path(remote):q}")
        except Exception:
            return f"Failed to access shared directory {remote} on remote host."
        remote_files = [
            x for x in remote_files.splitlines() if x not in (".", "..")
        ]
        #
        if sorted(local_files) != sorted(remote_files):
            return f"shared directory {local} has different content on remote host under {remote}"

    return f'OK (shared {" ".join(host.shared_dirs)})'


def stty_sane():
    try:
        subprocess.check_call("stty sane", shell=True)
    except Exception:
        pass


def test_queue(host, cmd=None, verbosity=1):
    try:
        h = Host(host, start_engine=False)
    except Exception as e:
        if verbosity > 2:
            env.logger.warning(e)
        return [host, "?", "?", "-", "-", "-", "-", "-"]
    ssh_res = test_ssh(h._host_agent)
    return [
        h.alias,
        h._host_agent.address,
        h._engine_type,
        ssh_res,
        test_scp(h._host_agent) if ssh_res.startswith("OK") else "-",
        test_cmd(h._host_agent, [h.config.get("sos", "sos"), "-h"])
        if ssh_res.startswith("OK") else "-",
        test_paths(h._host_agent) if ssh_res.startswith("OK") else "-",
        test_shared(h._host_agent) if ssh_res.startswith("OK") else "-",
    ] + ([] if cmd is None else [test_cmd(h._host_agent, cmd)])


def test_queues(cfg, hosts=[], cmd=None, verbosity=1):
    env.verbosity = verbosity
    all_hosts = cfg.get("hosts", [])
    if not all_hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    host_description = [
        [
            "Alias", "Address", "Queue Type", "ssh", "scp", "sos", "paths",
            "shared"
        ] + ([] if cmd is None else [" ".join(cmd)]),
        [
            "-----", "-------", "----------", "---", "---", "---", "-----",
            "------"
        ] + ([] if cmd is None else ["-" * len(" ".join(cmd))]),
    ]
    for host in hosts:
        if host not in all_hosts:
            env.logger.warning(f"Undefined host {host}")
    hosts = [x for x in hosts if x in all_hosts] if hosts else all_hosts
    if not hosts:
        return
    from multiprocessing import Pool

    pool = Pool(min(len(hosts), 10))
    host_description.extend(
        pool.starmap(test_queue, [(x, cmd, verbosity) for x in hosts]))
    if verbosity == 0:
        # just print succ or self
        for hd in host_description:
            print(f'{hd[0]} {"OK" if all(x=="OK" for x in hd[3:]) else "FAIL"}')
    elif verbosity in (1, 2):
        shortened = host_description[:2]
        for row in host_description[2:]:
            shortened.append(row[:3] + [
                "OK" if x.startswith("OK") else ("-" if x == "-" else "FAIL")
                for x in row[3:]
            ])
        width = [(len(x) for x in row) for row in shortened]
        max_width = [max(x) for x in zip(*width)]
        print("\n".join(" ".join([t.ljust(w)
                                  for t, w in zip(row, max_width)])
                        for row in shortened))
        if any("FAILED" in row for row in shortened):
            print(
                '\nUse command "sos remote --test host -v3" to check details of hosts with failed tests.'
            )
    else:
        for row in host_description[2:]:
            print(f"Alias:       {row[0]}")
            print(f"Address:     {row[1]}")
            print(f"Queue Type:  {row[2]}")
            print(f"ssh:         {row[3]}")
            print(f"scp:         {row[4]}")
            print(f"sos:         {row[5]}")
            print(f"paths:       {row[6]}")
            print(f"shared:      {row[7]}")
            if cmd:
                print(
                    f'{" ".join(cmd)}:{" "*(max(1, 12 - len(" ".join(cmd))))}{row[8]}'
                )
            print()


def copy_public_key(host, agent, password):
    try:
        if password is None:
            import getpass

            password = getpass.getpass(
                f"Please enter password for {agent.address}: ")
        cmd = f"scp -P {agent.port if agent.port else 22} {os.path.expanduser('~')}/.ssh/id_rsa.pub {agent.address}:id_rsa.pub.{host}"
        env.logger.info(cmd)
        p = pexpect.spawn(cmd, echo=False)
        i = p.expect([
            "(?i)are you sure you want to continue connecting",
            "[pP]assword:",
            pexpect.EOF,
        ])
        if i == 0:
            p.sendline("yes")
            p.expect(
                [
                    "(?i)are you sure you want to continue connecting",
                    "[pP]assword:",
                    pexpect.EOF,
                ],
                timeout=5,
            )

        if i == 1:
            p.sendline(password)
            i = p.expect(["assword:", pexpect.EOF])
            if i == 0:
                p.close(force=True)
                return "Incorrect password specified (you can try to specify it with command line option --password)"
        if i == 2:
            p.close()
            return f"Failed to copy public key to {agent.address}"
    except Exception as e:
        p.close()
        return f"Failed to copy public key to {host}: {e}"
    #
    # ssh
    try:
        cmd = f"ssh {agent.address} -p {agent.port} '[ -d .ssh ] || mkdir .ssh && chmod 700 .ssh; cat id_rsa.pub.{host} >> .ssh/authorized_keys; rm -f id_rsa.pub.{host}'"
        env.logger.info(cmd)
        p = pexpect.spawn(cmd, echo=False)
        i = p.expect([
            "(?i)are you sure you want to continue connecting",
            "assword:",
            pexpect.EOF,
        ])
        if i == 0:
            p.sendline("yes")
            p.expect(
                [
                    "(?i)are you sure you want to continue connecting",
                    "[pP]assword:",
                    pexpect.EOF,
                ],
                timeout=5,
            )

        if i == 1:
            p.sendline(password)
            i = p.expect(["assword:", pexpect.EOF])
            if i == 0:
                p.close(force=True)
                return "Incorrect password specified (you can try to specify it with command line option --password)"
        elif i != 1:
            p.close()
            return "Failed to append public key to .ssh/authorized_keys"
    except Exception as e:
        p.close()
        return f"Failed to append public key to .ssh/authorized_keys: {e}"
    return "OK"


def create_public_key():
    try:
        cmd = "ssh-keygen -t rsa"
        env.logger.info(cmd)
        p = pexpect.spawn(cmd, echo=False)
        while True:
            i = p.expect([
                "Enter file in which to save .*",
                "Enter passphrase.*",
                "Enter same passphrase again:.*",
                "Overwrite .*",
                pexpect.EOF,
            ])
            if i in (0, 1, 2):
                p.sendline("")
            elif i == 3:
                p.sendline("y")
            elif i == 4:
                break
    except Exception as e:
        p.close()
        raise RuntimeError(f"Failed to create a public key: {e}")


def setup_remote_access(cfg, hosts=[], password="", verbosity=1):
    env.verbosity = verbosity
    all_hosts = cfg.get("hosts", [])
    if not all_hosts and not hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    for host in hosts:
        if host not in all_hosts:
            env.logger.warning(
                f"Treating undefined host {host} as address of a remote host.")
    # public_key
    public_key = os.path.join(os.path.expanduser("~"), ".ssh", "id_rsa.pub")

    from argparse import Namespace

    for host in sorted(hosts if hosts else all_hosts):
        try:
            if host in all_hosts:
                h = Host(host, start_engine=False, test_connection=False)
                host_agent = h._host_agent
                #
                # can ssh already?
                response = test_ssh(host_agent)
                if response.startswith("OK"):
                    env.logger.info(
                        f"Public key access is already enabled for host ``{host}`` with address ``{host_agent.address}``"
                    )
                    continue
                elif "Could not resolve hostname" in response:
                    env.logger.error(response)
                    sys.exit(1)
            else:
                host_agent = Namespace(address=host, port=22)
        except Exception as e:
            env.logger.error(
                f"Failed to start set up remote engine for {host}: {e}")
            continue

        if os.path.isfile(public_key):
            env.logger.info("Using existing public key .ssh/id_rsa.pub")
        else:
            env.logger.info(f"Public key {public_key} is found. Creating one.")
            create_public_key()
            if not os.path.isfile(public_key):
                raise RuntimeError("Failed to create public key.")
        #
        response = copy_public_key(host, host_agent, password)
        stty_sane()
        if not response.startswith("OK"):
            env.logger.error(response)
            sys.exit(1)
        # file copied, check ssh again.
        if isinstance(host_agent, Namespace):
            host_agent = Host(
                host, start_engine=False, test_connection=False)._host_agent
        response = test_ssh(host_agent)
        if response.startswith("OK"):
            env.logger.info(
                f"Public key access is successfully set up for host ``{host}`` with address ``{host_agent.address}``"
            )
            continue
        else:
            env.logger.error(
                f"Failed to connect to {host} after passing public key. Possible problems include permission of .ssh and home directories."
            )


def login_host(cfg, host):
    try:
        h = Host(host, start_engine=False)
    except Exception as e:
        raise ValueError(f"Unrecognized or invalid host {host}: {e}")

    address, port = h._host_agent.address, h._host_agent.port
    try:
        env.logger.info(f"Running ``ssh {address} -p {port}``")
        os.execvp("ssh", ["ssh", address, "-p", str(port)])
    except Exception as e:
        raise RuntimeError(f"Failed to log in to {host}: {e}")


def run_command_on_hosts(cfg, hosts, cmd, verbosity):
    env.verbosity = verbosity
    if not hosts:
        hosts = cfg.get("hosts", [])
    if not hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    for host in hosts:
        # runing command on all hosts
        try:
            env.logger.info(f'Running ``{" ".join(cmd)}`` on ``{host}``')
            h = Host(host, start_engine=False)
            print(h._host_agent.check_output(cmd, under_workdir=True))
        except Exception as e:
            from .utils import get_traceback

            if verbosity and verbosity > 2:
                sys.stderr.write(get_traceback())
            env.logger.error(str(e))


def push_to_hosts(cfg, hosts, items, verbosity):
    env.verbosity = verbosity
    if not hosts:
        hosts = cfg.get("hosts", [])
    if not hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    for host in hosts:
        try:
            env.logger.info(f'Pushing ``{" ".join(items)}`` to ``{host}``')

            h = Host(host, start_engine=False)
            #
            sent = h.send_to_host(items)
            #
            env.logger.info("{} item{} sent:\n{}".format(
                len(sent),
                " is" if len(sent) <= 1 else "s are",
                "\n".join([
                    "{} => {}".format(x, sent[x]) for x in sorted(sent.keys())
                ]),
            ))
        except Exception as e:
            from .utils import get_traceback

            if verbosity and verbosity > 2:
                sys.stderr.write(get_traceback())
            env.logger.error(str(e))
            sys.exit(1)


def pull_from_host(cfg, hosts, items, verbosity):
    env.verbosity = verbosity
    if not hosts:
        hosts = cfg.get("hosts", [])
    if not hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    if len(hosts) > 1:
        raise ValueError("Can only pull from a single remote host.")
    try:
        env.logger.info(f'Pulling ``{" ".join(items)}`` from ``{hosts[0]}``')

        host = Host(hosts[0], start_engine=False)
        #
        received = host.receive_from_host(items)
        #
        print("{} item{} received:\n{}".format(
            len(received),
            " is" if len(received) <= 1 else "s are",
            "\n".join([
                "{} <= {}".format(x, received[x])
                for x in sorted(received.keys())
            ]),
        ))
    except Exception as e:
        from .utils import get_traceback

        if verbosity and verbosity > 2:
            sys.stderr.write(get_traceback())
        env.logger.error(str(e))
        sys.exit(1)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import collections
import copy
import os
import re
import sys
from itertools import chain
from typing import Any, Dict, List, Optional, Union

from .syntax import SOS_WILDCARD
from .utils import env

__all__ = ["expand_pattern"]

#
# The following is adapted from snakemake so that the syntax of the pattern parameter
# would be the same as snakemake.
#
# https://bitbucket.org/snakemake/snakemake/src/22eff35627401d3bb243e068c5dce97107e7090b/snakemake/io.py?at=master&fileviewer=file-view-default
#
# I will re-implement it if there is any license issue with the code
#


def regex(filepattern: str) -> str:
    f = []
    last = 0
    wildcards = set()
    for match in SOS_WILDCARD.finditer(filepattern):
        f.append(re.escape(filepattern[last:match.start()]))
        wildcard = match.group("name")
        if wildcard in wildcards:
            if match.group("constraint"):
                raise ValueError(
                    "If multiple wildcards of the same name "
                    "appear in a string, eventual constraints have to be defined "
                    "at the first occurence and will be inherited by the others."
                )
            f.append(f"(?P={wildcard})")
        else:
            wildcards.add(wildcard)
            f.append("(?P<{}>{})".format(
                wildcard,
                match.group("constraint")
                if match.group("constraint") else ".+",
            ))
        last = match.end()
    f.append(re.escape(filepattern[last:]))
    f.append("$")  # ensure that the match spans the whole file
    return "".join(f)


def glob_wildcards(
    pattern: str,
    files: Optional[List[str]] = None
) -> Dict[str, Union[List[Any], List[str]]]:
    """
    Glob the values of the wildcards by matching the given pattern to the filesystem.
    Returns a named tuple with a list of values for each wildcard.
    """
    pattern = os.path.normpath(pattern)
    if sys.platform == "win32":
        # we perform path matching with / slash only
        pattern = pattern.replace("\\", "/")
    first_wildcard = re.search("{[^{]", pattern)
    dirname = (
        os.path.dirname(pattern[:first_wildcard.start()])
        if first_wildcard else os.path.dirname(pattern))
    if not dirname:
        dirname = "."

    names = [match.group("name") for match in SOS_WILDCARD.finditer(pattern)]
    res = {x: [] for x in names}
    pattern = re.compile(regex(pattern))

    if files is None:
        files = ((os.path.join(dirpath, f) if dirpath != "." else f)
                 for dirpath, dirnames, filenames in os.walk(dirname)
                 for f in chain(filenames, dirnames))

    for f in files:
        # we perform path matching with only / slash
        match = re.match(pattern, str(f).replace("\\", "/"))
        if match:
            for name, value in match.groupdict().items():
                res[name].append(value)
    return res


def apply_wildcards(
    pattern: str,
    wildcards: Dict[str, Union[int, str]],
    fill_missing: bool = False,
    fail_dynamic: bool = False,
    dynamic_fill: None = None,
    keep_dynamic: bool = False,
) -> str:

    def format_match(match):
        name = match.group("name")
        try:
            value = wildcards[name]
            if fail_dynamic and value == dynamic_fill:
                raise RuntimeError(name)
            return str(value)  # convert anything into a str
        except KeyError as ex:
            if keep_dynamic:
                return f"{{{name}}}"
            elif fill_missing:
                return dynamic_fill
            else:
                raise RuntimeError(f"Wildcard apply error: {ex} ({wildcards})")

    return SOS_WILDCARD.sub(format_match, pattern)


def extract_pattern(pattern: str, ifiles: List[str]) -> Dict[str, any]:
    """This function match pattern to a list of input files, extract and return
    pieces of filenames as a list of variables with keys defined by pattern."""
    res = glob_wildcards(pattern, [])
    for ifile in ifiles:
        matched = glob_wildcards(pattern, [ifile])
        for key in matched.keys():
            if not matched[key]:
                # env.logger.warning('Filename {} does not match pattern {}. None returned.'.format(ifile, pattern))
                res[key].append(None)
            else:
                res[key].extend(matched[key])
    return res


def expand_pattern(pattern: str) -> List[str]:
    """This function expand patterns against the current namespace
    and return a list of filenames"""
    ofiles = []
    sz = None
    res = glob_wildcards(pattern, [])
    sz = None
    wildcard = [{}]
    for key in res.keys():
        if key not in env.sos_dict:
            raise ValueError(f"Undefined variable {key} in pattern {pattern}")
        if not isinstance(env.sos_dict[key], str) and isinstance(
                env.sos_dict[key], collections.Sequence):
            if sz is None:
                sz = len(env.sos_dict[key])
                wildcard = [copy.deepcopy(wildcard[0]) for x in range(sz)]
            elif sz != len(env.sos_dict[key]):
                raise ValueError(
                    f"Variables in output pattern should have the same length (other={sz}, len({key})={len(env.sos_dict[key])})"
                )
            for idx, value in enumerate(env.sos_dict[key]):
                wildcard[idx][key] = value
        else:
            for v in wildcard:
                v[key] = env.sos_dict[key]
    #
    for card in wildcard:
        ofiles.append(
            apply_wildcards(
                pattern,
                card,
                fill_missing=False,
                fail_dynamic=False,
                dynamic_fill=None,
                keep_dynamic=False,
            ))
    return ofiles



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil
import subprocess
import sys
import time
import unittest
from contextlib import contextmanager

import pytest

from sos import execute_workflow
from sos.hosts import Host
from sos.parser import SoS_Script
from sos.targets import file_target
from sos.tasks import TaskFile, TaskParams
from sos.utils import env, textMD5
from sos.workflow_executor import Base_Executor

has_docker = sys.platform != "win32"
try:
    if sys.platform != "win32":
        subprocess.check_output(
            "docker ps | grep test_sos", shell=True).decode()
except subprocess.CalledProcessError:
    subprocess.call("sh build_test_docker.sh", shell=True)
    try:
        subprocess.check_output(
            "docker ps | grep test_sos", shell=True).decode()
    except subprocess.CalledProcessError:
        print("Failed to set up a docker machine with sos")
        has_docker = False


@contextmanager
def cd_new(path):
    old_dir = os.getcwd()
    if os.path.isdir(path):
        shutil.rmtree(path)
    os.makedirs(path, exist_ok=True)
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


def get_tasks():
    from sos.signatures import WorkflowSignatures
    env.exec_dir = os.path.join(
        os.path.expanduser("~"), ".sos", textMD5(os.getcwd()))
    db = WorkflowSignatures()
    conn = db.conn
    #conn = sqlite3.connect(os.path.join(env.exec_dir, "workflow_signatures.db"))
    cur = conn.cursor()
    cur.execute('SELECT DISTINCT id FROM workflows WHERE entry_type = "task"')
    return [x[0] for x in cur.fetchall()]


class TestTask(unittest.TestCase):

    def setUp(self):
        env.reset()
        subprocess.call("sos remove -s", shell=True)
        # self.resetDir('~/.sos')
        self.temp_files = []
        Host.reset()

    def touch(self, files):
        """create temporary files"""
        if isinstance(files, str):
            files = [files]
        #
        for f in files:
            with open(f, "w") as tmp:
                tmp.write("test")
        #
        self.temp_files.extend(files)

    def test_task_file(self):
        """Test task file handling"""
        for ext in (".pulse", ".out", ".err", ".task", ".sh"):
            filename = os.path.join(
                os.path.expanduser("~"), ".sos", "tasks",
                "ffffffffffffffff" + ext)
            if os.path.isfile(filename):
                os.remove(filename)
        params = TaskParams(
            name="ffffffffffffffff",
            global_def=None,
            task="b=a",
            sos_dict={"a": 1},
            tags=["b", "a"],
        )
        a = TaskFile("ffffffffffffffff")
        a.save(params)
        self.assertEqual(a.tags, "a b")
        for ext in (".pulse", ".out", ".err", ".sh"):
            with open(
                    os.path.join(
                        os.path.expanduser("~"), ".sos", "tasks",
                        "ffffffffffffffff" + ext),
                    "w",
            ) as fh:
                fh.write(ext)
        self.assertFalse(a.has_stdout())
        self.assertFalse(a.has_stderr())
        a.add_outputs()
        #
        self.assertEqual(a.params.sos_dict["a"], 1)
        self.assertEqual(a.status, "new")
        a.status = "completed"
        self.assertLess(time.time() - a.last_updated, 2)
        self.assertEqual(a.status, "completed")
        #
        # get and reset info
        info = a.info
        a.status = "running"
        self.assertEqual(a.status, "running")
        a.info = info
        self.assertEqual(a.status, "completed")
        self.assertTrue(a.has_stdout())
        #
        a.add_result({"ret_code": 5})
        #
        a.tags = ["ee", "d"]
        self.assertEqual(a.tags, "d ee")
        # a.add_tags(['kk'])
        # self.assertEqual(a.tags.split(), ['d', 'ee', 'kk'])
        #
        self.assertEqual(a.params.sos_dict["a"], 1)
        self.assertEqual(a.params.task, "b=a")
        #

        self.assertEqual(a.stdout, ".out")
        self.assertEqual(a.stderr, ".err")
        self.assertEqual(a.pulse, ".pulse")
        self.assertEqual(a.shell, ".sh")
        self.assertTrue(a.has_stdout())
        self.assertTrue(a.has_stderr())
        self.assertTrue(a.has_pulse())
        self.assertTrue(a.has_shell())
        #
        #
        a.reset()
        self.assertEqual(a.status, "new")
        self.assertEqual(a.stdout, "")
        self.assertEqual(a.stderr, "")
        self.assertEqual(a.signature, {})
        a.add_outputs()
        a.add_result({"ret_code": 5})
        self.assertEqual(a.result["ret_code"], 5)

    def test_workdir(self):
        """Test workdir option for runtime environment"""
        import tempfile

        tdir = tempfile.mkdtemp()
        with open(os.path.join(tdir, "aaa.pp"), "w") as aaa:
            aaa.write("something")
        script = r"""
import os
[0]
task: workdir={0!r}

with open(os.path.join({1!r}, 'result.txt'), 'w') as res:
   for file in os.listdir({1!r}):
       res.write(file + '\n')
""".format(os.path.split(tdir)[0],
           os.path.split(tdir)[1])
        wf = SoS_Script(script).workflow()
        env.config["sig_mode"] = "force"
        Base_Executor(wf, config={"default_queue": "localhost"}).run()
        with open(os.path.join(tdir, "result.txt")) as res:
            content = [x.strip() for x in res.readlines()]
            self.assertTrue("aaa.pp" in content)

    #     def testSequential(self):
    #         '''Test concurrency option for runtime environment'''
    #         env.max_jobs = 5
    #         start = time.time()
    #         execute_workflow(
    #             script=r"""
    # import time
    # [0]

    # repeat = range(4)
    # input: for_each='repeat'

    # task: concurrent=False

    # print('I am {}, waited {} seconds'.format(_index, _repeat + 1))
    # time.sleep(_repeat + 1)
    # print('I am {}, done'.format(_index))
    # """,
    #             options={
    #                 'default_queue': 'localhost',
    #                 'sig_mode': 'force'
    #             })
    #         self.assertGreater(time.time() - start, 11)

    def test_concurrency(self):
        """Test concurrency option for runtime environment"""
        execute_workflow(
            r"""
[0]

repeat = range(4)
input: for_each='repeat'

task:

import time
print('I am {}, waited {} seconds'.format(_index, _repeat + 1))
time.sleep(_repeat + 1)
print('I am {}, done'.format(_index))
""",
            options={
                "default_queue": "localhost",
                "sig_mode": "force"
            },
        )

    def test_prepend_path(self):
        """Test prepend path"""
        import stat

        if not os.path.isdir("temp"):
            os.mkdir("temp")
        if sys.platform == "win32":
            with open(r"temp\temp_cmd.bat", "w") as tc:
                tc.write('echo "a"')
        else:
            with open("temp/temp_cmd", "w") as tc:
                tc.write('echo "a"')
            os.chmod("temp/temp_cmd",
                     stat.S_IXUSR | stat.S_IWUSR | stat.S_IRUSR)
        #
        script = SoS_Script(r"""
[1]
task:
run:
    temp_cmd
""")
        script.workflow()
        env.config["sig_mode"] = "force"
        # self.assertRaises(Exception, Base_Executor(wf).run)
        #
        # the following is supposed to create its own task file but
        # for some reason it uses the same task file
        #
        # use option env
        execute_workflow(
            script=r"""
import os
[1]
task: env={'PATH': 'temp' + os.pathsep + os.environ['PATH']}
run:
    temp_cmd
""",
            options={
                "default_queue": "localhost",
                "sig_mode": "force"
            },
        )
        #
        #
        execute_workflow(
            script=r"""
[1]
task: prepend_path='temp'
run:
    temp_cmd
""",
            options={"default_queue": "localhost"},
        )

    def test_no_task(self):
        env.config["sig_mode"] = "force"
        script = SoS_Script(r"""

[10]
task:
run:
   sleep 0
""")
        wf = script.workflow()
        # this will always work and not through task
        Base_Executor(wf, config={"default_queue": None}).run()
        #
        env.config["sig_mode"] = "force"
        script = SoS_Script(r"""

[10]
task: queue=None
run:
   sleep 0
""")
        wf = script.workflow()
        # this will always work and not through task
        Base_Executor(wf, config={"default_queue": "localhost"}).run()

    def test_passing_var_to_task(self):
        """Test passing used variable to tasks"""
        for i in range(10, 13):
            if file_target(f"myfile_{i}.txt").exists():
                file_target(f"myfile_{i}.txt").unlink()
        #
        env.config["sig_mode"] = "force"
        script = SoS_Script(r"""
parameter: gvar = 10

[10]
# generate a file
tt = range(gvar, gvar + 3)
input: for_each='tt'
output: f"myfile_{_tt}.txt"
# additional comment

# _tt should be used in task
task:
run: expand=True
    echo {_tt}_{_index} > {_output:q}

""")
        wf = script.workflow()
        env.max_jobs = 4
        Base_Executor(wf, config={"default_queue": "localhost"}).run()
        for t in range(10, 13):
            with open(f"myfile_{t}.txt") as tmp:
                self.assertEqual(tmp.read().strip(), str(t) + "_" + str(t - 10))
            if file_target(f"myfile_{t}.txt").exists():
                file_target(f"myfile_{t}.txt").unlink()

    def test_max_jobs(self):
        """Test default max number of jobs"""
        script = SoS_Script(r"""

[10]
input: for_each=[{'a': range(2)}, {'b': range(3)}]

task:
run: expand=True
    echo "a = {a}, b = {b}"
""")
        env.config["max_running_jobs"] = 2
        wf = script.workflow()
        Base_Executor(wf, config={"default_queue": "localhost"}).run()

    def test_kill_and_purge(self):
        """Test no wait"""
        subprocess.call(["sos", "purge"])
        with open("test_purge.sos", "w") as script:
            script.write(r"""
[10]
task:
run:
    echo Try to kill
    sleep 20
""")
        subprocess.Popen(
            "sos run test_purge.sos -s force -q localhost", shell=True)
        time.sleep(5)
        subprocess.call(["sos", "kill", "--all"])
        for i in range(20):
            output = subprocess.check_output(["sos", "status", "-v",
                                              "1"]).decode()
            if "killed" in output or "aborted" in output or "completed" in output:
                break
            self.assertFalse(
                i > 10,
                "Task should be killed within 10 seconds, got {}".format(
                    output))
            time.sleep(1)
        # test purge by status
        subprocess.call(["sos", "purge", "--status", "aborted"])
        self.assertFalse("killed" in subprocess.check_output(
            ["sos", "status", "-v", "3"]).decode())
        # purge by all is not tested because it is dangerous

    def test_concurrent_task(self):
        """Test submitting tasks from concurrent substeps"""
        for f in [f"con_{x}.txt" for x in range(5)]:
            if file_target(f).exists():
                file_target(f).unlink()
        script = SoS_Script("""
[10]
input: for_each={'i': range(5)}
output: f'con_{i}.txt'

task:
run: expand=True
  echo {i} > {_output}
""")
        wf = script.workflow()
        Base_Executor(
            wf, config={
                "sig_mode": "force",
                "default_queue": "localhost"
            }).run()
        for f in [f"con_{x}.txt" for x in range(5)]:
            self.assertTrue(file_target(f).exists())

    def test_shared_option(self):
        """Test shared option of task"""
        for f in ("a.txt", "a100.txt"):
            if file_target(f).exists():
                file_target(f).unlink()
        script = SoS_Script("""
[10: shared = 'a']
output: 'a.txt'
task: shared={'a': 'int(open("a.txt").read())'}
run:
  echo 100 > a.txt

[20]
run: expand=True
    touch a{a}.txt
""")
        wf = script.workflow()
        Base_Executor(
            wf, config={
                "sig_mode": "force",
                "default_queue": "localhost"
            }).run()
        self.assertTrue(os.path.isfile("a100.txt"))
        # sequence of var or mapping
        for f in ("a.txt", "a100.txt"):
            if file_target(f).exists():
                file_target(f).unlink()
        script = SoS_Script("""
[10: shared = ['a', 'b']]
output: 'a.txt'
task: shared=[{'a': 'int(open("a.txt").read())'}, 'b']
b = 20
run:
  echo 100 > a.txt

[20]
run: expand=True
    touch a{a}_{b}.txt
""")
        wf = script.workflow()
        Base_Executor(
            wf, config={
                "sig_mode": "force",
                "default_queue": "localhost"
            }).run()
        self.assertTrue(os.path.isfile("a100_20.txt"))

        script = SoS_Script("""
[10 (simulate): shared=['rng', 'step_rng']]
input: for_each={'i': range(5)}
task: shared='rng'
print(f"{i}")
import random
rng = random.randint(1, 1000)
""")
        wf = script.workflow()
        Base_Executor(wf, config={"default_queue": "localhost"}).run()
        var = env.sos_dict["rng"]
        self.assertTrue(isinstance(var, int))
        self.assertTrue(isinstance(env.sos_dict["step_rng"], list))
        self.assertEqual(env.sos_dict["step_rng"][-1], var)
        # run it again, should get from signature
        #
        # Base_Executor(wf).run()
        # self.assertEqual(var, env.sos_dict['rng'])

    def test_trunk_size_option(self):
        """Test option trunk_size"""
        with open("test_trunksize.sos", "w") as tt:
            tt.write("""
[10]
input: for_each={'I': range(10)}
task: trunk_size=5, cores=1, mem='1M', walltime='10m'
run: expand=True
    echo {I} > {I}.txt
    sleep 0.1
""")
        wf = SoS_Script(filename="test_trunksize.sos").workflow()
        for i in range(10):
            if os.path.isfile(f"{i}.txt"):
                file_target(f"{i}.txt").unlink()
        Base_Executor(
            wf,
            config={
                "sig_mode": "force",
                "script": "test_trunksize.sos",
                "max_running_jobs": 10,
                "workflow_args": [],
                "output_dag": "",
                "output_report": None,
                "targets": [],
                "worker_procs": ["4"],
                "default_queue": "localhost",
                "workflow": "default",
                "workdir": ".",
            },
        ).run()

        for i in range(10):
            self.assertTrue(os.path.isfile(f"{i}.txt"))
            os.remove(f"{i}.txt")
        # trunk size is None or 0, -1, intepreted as all tasks
        with open("test_trunksize.sos", "w") as tt:
            tt.write("""
[10]
input: for_each={'I': range(10)}
task: trunk_size=None, cores=1, mem='1M', walltime='10m'
run: expand=True
    echo {I} > {I}.txt
    sleep 0.1
""")
        wf = SoS_Script(filename="test_trunksize.sos").workflow()
        Base_Executor(
            wf,
            config={
                "sig_mode": "force",
                "script": "test_trunksize.sos",
                "max_running_jobs": 10,
                "workflow_args": [],
                "output_dag": "",
                "output_report": None,
                "targets": [],
                "worker_procs": ["4"],
                "default_queue": "localhost",
                "workflow": "default",
                "workdir": ".",
            },
        ).run()
        for i in range(10):
            self.assertTrue(os.path.isfile(f"{i}.txt"))

    def test_trunk_workers_option(self):
        """Test option trunk_workers"""
        with open("test_trunkworker.sos", "w") as tt:
            tt.write("""
[10]
input: for_each={'I': range(12)}
task: trunk_size=6, trunk_workers=3, mem='1M', walltime='10m'
run: expand=True
    echo {I} > {I}.txt
    sleep 2
""")
        wf = SoS_Script(filename="test_trunkworker.sos").workflow()
        Base_Executor(
            wf,
            config={
                "sig_mode": "force",
                "script": "test_trunkworker.sos",
                "max_running_jobs": 10,
                "workflow_args": [],
                "output_dag": "",
                "output_report": None,
                "targets": [],
                "worker_procs": ["4"],
                "default_queue": "localhost",
                "workflow": "default",
                "workdir": ".",
            },
        ).run()
        for i in range(10):
            self.assertTrue(os.path.isfile("{}.txt".format(i)))

    def test_task_tags(self):
        """Test option tags of tasks"""
        import random

        tag = "tag{}".format(random.randint(1, 100000))
        with open("test_tags.sos", "w") as tt:
            tt.write("""
[10]
input: for_each={{'i': range(10)}}
task: tags='{}', trunk_size=2
sh: expand=True
  echo {} {{i}}
""".format(tag, tag))
        wf = SoS_Script(filename="test_tags.sos").workflow()
        Base_Executor(
            wf,
            config={
                "sig_mode": "force",
                "script": "test_trunkworker.sos",
                "max_running_jobs": 10,
                "workflow_args": [],
                "output_dag": "",
                "output_report": None,
                "default_queue": "localhost",
                "targets": [],
                "worker_procs": ["4"],
                "workflow": "default",
                "workdir": ".",
            },
        ).run()
        ret = subprocess.check_output(
            "sos status -t {}".format(tag), shell=True).decode()
        self.assertEqual(len(ret.splitlines()), 5, "Obtained {}".format(ret))
        # test multiple tags
        tag1 = "tag{}".format(random.randint(1, 100000))
        tag2 = "tag{}".format(random.randint(1, 100000))
        with open("test_tags.sos", "w") as tt:
            tt.write("""
[10]
input: for_each={{'i': range(2)}}
task: tags=['{}', '{}']
sh: expand=True
  echo {} {{i}}
""".format(tag1, tag2, tag1))
        wf = SoS_Script(filename="test_tags.sos").workflow()
        Base_Executor(
            wf,
            config={
                "sig_mode": "force",
                "script": "test_trunkworker.sos",
                "max_running_jobs": 10,
                "workflow_args": [],
                "output_dag": "",
                "output_report": None,
                "targets": [],
                "worker_procs": ["4"],
                "default_queue": "localhost",
                "workflow": "default",
                "workdir": ".",
            },
        ).run()
        ret = subprocess.check_output(
            "sos status -t {}".format(tag2), shell=True).decode()
        self.assertEqual(len(ret.splitlines()), 2, "Obtained {}".format(ret))

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_max_mem(self):
        """Test server restriction max_mem"""
        script = SoS_Script("""
[10]
task: mem='2G'
print('a')
""")
        wf = script.workflow()
        self.assertRaises(
            Exception,
            Base_Executor(
                wf,
                config={
                    "config_file": "~/docker.yml",
                    "default_queue": "docker_limited",
                    "sig_mode": "force",
                },
            ).run,
        )

    def test_local_runtime_max_walltime(self):
        """Test server max_walltime option"""
        script = SoS_Script("""
[10]
task:
import time
time.sleep(15)
""")
        wf = script.workflow()
        self.assertRaises(
            Exception,
            Base_Executor(
                wf,
                config={
                    "config_file": "~/docker.yml",
                    "default_queue": "local_limited",
                    "sig_mode": "force",
                },
            ).run,
        )

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_max_cores(self):
        """Test server restriction max_cores"""
        script = SoS_Script("""
[10]
task: cores=8
print('a')
""")
        wf = script.workflow()
        self.assertRaises(
            Exception,
            Base_Executor(
                wf,
                config={
                    "config_file": "~/docker.yml",
                    "default_queue": "docker_limited",
                    "sig_mode": "force",
                },
            ).run,
        )

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_override_max_cores(self):
        """Test use queue_args to override server restriction max_cores"""
        script = SoS_Script("""
[10]
task: cores=8
print('a')
""")
        wf = script.workflow()
        Base_Executor(
            wf,
            config={
                "config_file": "~/docker.yml",
                "default_queue": "docker_limited",
                "sig_mode": "force",
                "queue_args": {
                    "cores": 1
                },
            },
        ).run()

    def test_list_hosts(self):
        """test list hosts using sos status -q"""
        for v in ["0", "1", "3", "4"]:
            output = subprocess.check_output(
                ["sos", "remote", "list", "-c", "~/docker.yml", "-v",
                 v]).decode()
            self.assertTrue(
                "local_limited" in output,
                f"local_limited not in \n{output}\n for verbosity {v}",
            )

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_max_walltime(self):
        """Test server restriction max_walltime"""
        script = SoS_Script("""
[10]
task: walltime='1:00:00'
print('a')
""")
        wf = script.workflow()
        self.assertRaises(
            Exception,
            Base_Executor(
                wf,
                config={
                    "config_file": "~/docker.yml",
                    "default_queue": "docker_limited",
                    "sig_mode": "force",
                },
            ).run,
        )

    def test_purge_all_with_option(self):
        """Test sos purge all with options such as -s completed"""
        with cd_new("temp_c"):
            with open("test.sos", "w") as tst:
                tst.write("""
input: for_each={'i': range(2)}
output: f'a{i}.txt'
task:
run: expand=True
echo temp_a
touch {_output}
""")
            subprocess.call("sos run test -s force -q localhost", shell=True)
            tasks = get_tasks()
            subprocess.call("sos purge -s failed", shell=True)
        # check tasks
        taskstatus = [
            x.split()[0] for x in subprocess.check_output(
                "sos status -v1", shell=True).decode().splitlines()
        ]
        self.assertTrue(all(x in taskstatus for x in tasks))
        # purge one of them
        subprocess.call(f"sos purge {tasks[0]}", shell=True)
        taskstatus = [
            x.split()[0] for x in subprocess.check_output(
                "sos status -v1", shell=True).decode().splitlines()
        ]
        self.assertTrue(tasks[0] not in taskstatus)
        self.assertTrue(tasks[1] in taskstatus)
        #
        subprocess.call("sos purge --all", shell=True)
        taskstatus = [
            x.split()[0] for x in subprocess.check_output(
                "sos status -v1", shell=True).decode().splitlines()
        ]
        self.assertTrue(tasks[1] not in taskstatus)

    def test_resubmit_task_with_different_walltime(self):
        """Test resubmission of tasks with different walltime #1019"""
        with cd_new("temp_walltime"):
            with open("test.sos", "w") as tst:
                tst.write("""
task: walltime='1m'
sh:
echo 0.1
""")
            subprocess.call("sos run test -s force -q localhost", shell=True)
            tasks = get_tasks()
            out = subprocess.check_output(
                f"sos status {tasks[0]} -v4", shell=True)
            self.assertTrue("00:01:00" in out.decode())
            with open("test1.sos", "w") as tst:
                tst.write("""
task: walltime='2m'
sh:
echo 0.1
""")
            subprocess.call("sos run test1 -s force -q localhost", shell=True)
            new_tasks = get_tasks()
            self.assertEqual(tasks, new_tasks)
            #
            out = subprocess.check_output(
                f"sos status {tasks[0]} -v4", shell=True)
            self.assertTrue("00:02:00" in out.decode())

    def test_task_signature(self):
        """Test re-execution of tasks"""
        with cd_new("temp_signature"):
            with open("test.sos", "w") as tst:
                tst.write("""
task:
sh:
sleep 2
""")
            subprocess.call("sos run test -s force -q localhost", shell=True)
            tasks = get_tasks()
            tf = TaskFile(tasks[0])
            self.assertTrue(tf.has_signature())
            self.assertEqual(tf.status, "completed")
            tf.tags_created_start_and_duration()
            #
            subprocess.call("sos run test -q localhost", shell=True)
            self.assertLess(tf.tags_created_start_and_duration()[3], 1)

    def test_wrong_host(self):
        script = SoS_Script("""
[10]
task: walltime='1:00:00', queue='undefined'
print('a')
""")
        wf = script.workflow()
        self.assertRaises(
            Exception,
            Base_Executor(wf, config={
                "default_queue": "localhost"
            }).run)

    def test_output_in_task(self):
        """Test passing _output to task #1136"""
        script = SoS_Script("""
chunks  = [1,2]
[1]
input: for_each = 'chunks'
output: f'{_chunks}.txt'
_output.touch()

[2]
input: group_with = 'chunks'
output: summary_stats = f'{_input}.summary', ld_matrix = f'{_input}.result'
task:

python3: expand="${ }"
       open("${_output['summary_stats']}", 'w').close()
       open("${_output['ld_matrix']}", 'w').close()
""")
        wf = script.workflow()
        Base_Executor(wf, config={"default_queue": "localhost"}).run()

    def test_repeated_tasks(self):
        """Test statement before task #1142 """
        script = SoS_Script("""
[1]
input: for_each=dict(i=range(5))

print(f'This is for {i}')
task:  walltime='10m'
print(f'this is task {i}')

""")
        for i in range(5):
            wf = script.workflow()
            Base_Executor(
                wf, config={
                    "sig_mode": "force",
                    "default_queue": "localhost"
                }).run()

    def test_passing_parameters_to_task(self):
        """Test passing of parameters in global section to tasks #1155"""
        script = SoS_Script("""\
[global]
parameter: par=5
def a():
  print(par)

[default]
task:
a()
""")
        wf = script.workflow()
        Base_Executor(
            wf, config={
                "sig_mode": "force",
                "default_queue": "localhost"
            }).run()

    def test_trunk_size_with_stop_if(self):
        """Test a case when some tasks are not submitted due to holes in slots #1159"""
        for i in range(5):
            f = f"{i+1}.txt"
            if os.path.isfile(f):
                os.remove(f)
        script = SoS_Script("""\
[1]
output: [f'{x+1}.txt' for x in range(5)]
for i in range(5):
  name = f'{i+1}.txt'
  if i not in [0,1,2]:
    path(name).touch()
  else:
    with open(name, 'w') as f:
      f.write('test it')

[2]
input: group_by = 1
output: f'{_input:n}.out'
stop_if(_input.stat().st_size==0, no_output=True)

task: trunk_size = 80
_output.touch()
""")
        wf = script.workflow()
        Base_Executor(
            wf, config={
                "sig_mode": "force",
                "default_queue": "localhost"
            }).run()

    def test_output_from_master_task(self):
        """Test splitting the output from master task #1203"""
        script = SoS_Script("""\
l=[x for x in range(1,13)]

[2]
input: for_each = 'l'
output: f'{_l}.out'

task: trunk_size = 4
_output.touch()

[3]
assert _input == f'{_index+1}.out'
""")
        wf = script.workflow()
        Base_Executor(wf, config={"default_queue": "localhost"}).run()

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_sync_input_output_and_rerun(self):
        """Test sync input and output with remote host"""
        for i in range(4):
            if os.path.isfile(f"test_{i}.txt"):
                os.remove(f"test_{i}.txt")
            if os.path.isfile(f"test_{i}.bak"):
                os.remove(f"test_{i}.bak")
        import random

        script = SoS_Script("""
parameter: g = 100

[10]
input: for_each=dict(i=range(4))
output: f'test_{i}.txt'

with open(f'test_{i}.txt', 'w') as tst:
    tst.write(f'test_{i}_{g}')

[20]
output: _input.with_suffix('.bak')

task:

with open(_input, 'r') as inf, open(_output, 'w') as outf:
	outf.write(inf.read() + '.bak')
""")
        wf = script.workflow()
        val = random.randint(1, 10000)
        Base_Executor(
            wf,
            args=["--g", str(val)],
            config={
                "config_file": "~/docker.yml",
                "default_queue": "docker",
                "sig_mode": "force",
            },
        ).run()
        # now check if
        for i in range(4):
            self.assertTrue(os.path.isfile(f"test_{i}.txt"))
            with open(f"test_{i}.bak") as outf:
                self.assertEqual(outf.read(), f"test_{i}_{val}.bak")
            self.assertTrue(os.path.isfile(f"test_{i}.bak"))
            with open(f"test_{i}.bak") as outf:
                self.assertEqual(outf.read(), f"test_{i}_{val}.bak")
        #
        # test rerun the task file on local host
        for i in range(4):
            if os.path.isfile(f"test_{i}.txt"):
                os.remove(f"test_{i}.txt")
            if os.path.isfile(f"test_{i}.bak"):
                os.remove(f"test_{i}.bak")
        Base_Executor(
            wf, args=["--g", str(val)], config={
                "sig_mode": "force"
            }).run()
        for i in range(4):
            self.assertTrue(os.path.isfile(f"test_{i}.txt"))
            with open(f"test_{i}.bak") as outf:
                self.assertEqual(outf.read(), f"test_{i}_{val}.bak")
            self.assertTrue(os.path.isfile(f"test_{i}.bak"))
            with open(f"test_{i}.bak") as outf:
                self.assertEqual(outf.read(), f"test_{i}_{val}.bak")

    @unittest.skipIf(not has_docker, "Docker container not usable")
    def test_remote_input_target(self):
        """Test the use of remote target"""
        if os.path.isfile("vars.sh"):
            os.remove("vars.sh")
        if os.path.isfile("vars1.sh"):
            os.remove("vars1.sh")
        script = SoS_Script("""

[10]
input: remote('/lib/init/vars.sh')
output: f'vars1.sh'

task:

with open(_input, 'r') as inf, open(_output, 'w') as outf:
	outf.write(inf.read())
""")
        wf = script.workflow()
        Base_Executor(
            wf,
            config={
                "config_file": "~/docker.yml",
                "default_queue": "docker",
                "sig_mode": "force",
            },
        ).run()
        self.assertFalse(os.path.isfile("vars.sh"))
        self.assertTrue(os.path.isfile("vars1.sh"))


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_delayed_interpolation(clear_now_and_after):
    """Test delayed interpolation with expression involving remote objects"""
    # purge all previous tasks
    clear_now_and_after('test.py', 'test.py.bak')
    execute_workflow(
        """
        [10]
        output: remote('test.py')
        task:
        run:
            touch test.py

        [20]
        output: remote(f"{_input:R}.bak")
        task:
        run: expand=True
            cp {_input} {_output}
        """,
        options={
            "config_file": "~/docker.yml",
            # do not wait for jobs
            "wait_for_task": True,
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    # this file is remote only
    assert not os.path.isfile("test.py")
    assert not os.path.isfile("test.py.bak")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_output_target(clear_now_and_after):
    """Test the use of remote target"""
    clear_now_and_after("vars.sh", "vars1.sh")
    execute_workflow(
        """
        [10]
        input: remote('/lib/init/vars.sh')
        output: remote(f'vars1.sh')

        task:

        with open(_input, 'r') as inf, open(_output, 'w') as outf:
            outf.write(inf.read())
        """,
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert not os.path.isfile("vars.sh")
    assert not os.path.isfile("vars1.sh")


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_remote_output_target_with_trunksize(clear_now_and_after):
    clear_now_and_after("vars.sh", "vars1.sh")

    execute_workflow(
        """\
        [10]
        import os
        input: remote('/lib/init/vars.sh'), remote('/lib/init/init-d-script'), group_by=1
        output: remote(os.path.basename(str(_input)))

        task: trunk_size=2

        with open(_input, 'r') as inf, open(_output, 'w') as outf:
            outf.write(inf.read())""",
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    assert not os.path.isfile("vars.sh")
    assert not os.path.isfile("init-d-script")


def test_runtime_max_walltime():
    """Test server max_walltime option"""
    with pytest.raises(Exception):
        execute_workflow(
            """
        [10]
        task:
        import time
        time.sleep(25)
        """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "docker_limited",
                "sig_mode": "force",
            },
        )


@pytest.mark.skipif(not has_docker, reason="Docker container not usable")
def test_sync_master_task(clear_now_and_after):
    """Test sync input and output with remote host with trunksize"""
    clear_now_and_after([f"test_{i}.txt" for i in range(4)],
                        [f"test_{i}.bak" for i in range(4)])
    import random

    val = random.randint(1, 10000)
    execute_workflow(
        r"""
        parameter: g = 100

        [10]
        input: for_each=dict(i=range(4))
        output: f'test_{i}.txt'

        with open(f'test_{i}.txt', 'w') as tst:
            tst.write(f'test_{i}_{g}')

        [20]
        output: _input.with_suffix('.bak')

        task: trunk_size=2

        with open(_input, 'r') as inf, open(_output, 'w') as outf:
            outf.write(inf.read() + '.bak')
        """,
        args=["--g", str(val)],
        options={
            "config_file": "~/docker.yml",
            "default_queue": "docker",
            "sig_mode": "force",
        },
    )
    # now check if
    for i in range(4):
        assert os.path.isfile(f"test_{i}.txt")
        with open(f"test_{i}.bak") as outf:
            assert outf.read() == f"test_{i}_{val}.bak"
        assert os.path.isfile(f"test_{i}.bak")
        with open(f"test_{i}.bak") as outf:
            assert outf.read() == f"test_{i}_{val}.bak"


def test_local_max_cores():
    """Test server restriction max_cores"""
    with pytest.raises(Exception):
        execute_workflow(
            """
        [10]
        task: cores=8
        print('a')
        """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "local_limited",
                "sig_mode": "force",
            },
        )


def test_local_max_mem():
    """Test server restriction max_mem"""
    with pytest.raises(Exception):
        execute_workflow(
            """
        [10]
        task: mem='2G'
        print('a')
        """,
            options={
                "config_file": "~/docker.yml",
                "default_queue": "local_limited",
                "sig_mode": "force",
            },
        )



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import sys

__all__ = ["__version__", "SOS_FULL_VERSION"]

_py_ver = sys.version_info
if _py_ver.major == 2 or (_py_ver.major == 3 and
                          (_py_ver.minor, _py_ver.micro) < (6, 0)):
    raise SystemError(
        "SOS requires Python 3.6 or higher. Please upgrade your Python {}.{}.{}"
        .format(_py_ver.major, _py_ver.minor, _py_ver.micro))

# version of the SoS language
__sos_version__ = "1.0"
# version of the sos command
__version__ = "0.22.7"
__py_version__ = "{}.{}.{}".format(_py_ver.major, _py_ver.minor, _py_ver.micro)

#
SOS_FULL_VERSION = "{} for Python {}.{}.{}".format(__version__, _py_ver.major,
                                                   _py_ver.minor, _py_ver.micro)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import os
import stat
import threading
import time

import psutil

from .tasks import TaskFile
from .utils import env, expand_time, format_HHMMSS


class TaskMonitor(threading.Thread):

    def __init__(
        self,
        task_id,
        monitor_interval,
        resource_monitor_interval,
        max_walltime=None,
        max_mem=None,
        max_procs=None,
        sos_dict={},
    ):
        threading.Thread.__init__(self)
        self.task_id = task_id
        self.pid = os.getpid()
        self.monitor_interval = monitor_interval
        self.resource_monitor_interval = max(
            resource_monitor_interval // monitor_interval, 1)
        self.daemon = True
        if max_walltime is None:
            self.max_walltime = None
        else:
            mwall = expand_time(max_walltime[0]) if max_walltime[0] else None
            wall = expand_time(max_walltime[1]) if max_walltime[1] else None
            if mwall is not None and wall is not None:
                self.max_walltime = min(wall, wall)
            elif mwall is not None:
                self.max_walltime = mwall
            else:
                self.max_walltime = wall
        #
        if max_mem is None:
            self.max_mem = None
        elif max_mem[0] is not None and max_mem[1] is not None:
            self.max_mem = min(max_mem[0], max_mem[1])
        elif max_mem[0] is not None:
            self.max_mem = max_mem[0]
        else:
            self.max_mem = max_mem[1]
        #
        if max_procs is None:
            self.max_procs = None
        elif max_procs[0] is not None and max_procs[1] is not None:
            self.max_procs = min(max_procs[0], max_procs[1])
        elif max_procs[0] is not None:
            self.max_procs = max_procs[0]
        else:
            self.max_procs = max_procs[1]

        self.pulse_file = os.path.join(
            os.path.expanduser("~"), ".sos", "tasks", task_id + ".pulse")
        # remove previous status file, which could be readonly if the job is killed
        if os.path.isfile(self.pulse_file):
            if not os.access(self.pulse_file, os.W_OK):
                os.chmod(self.pulse_file, stat.S_IREAD | stat.S_IWRITE)
            os.remove(self.pulse_file)
        self.sos_dict = sos_dict
        with open(self.pulse_file, "a") as pd:
            pd.write(
                "#time\tproc_cpu\tproc_mem\tchildren\tchildren_cpu\tchildren_mem\n"
            )

    def _check(self):
        current_process = psutil.Process(self.pid)
        par_cpu = current_process.cpu_percent()
        par_mem = current_process.memory_info()[0]
        ch_cpu = 0
        ch_mem = 0
        children = current_process.children(recursive=True)
        n_children = len(children)
        for child in children:
            ch_cpu += child.cpu_percent()
            ch_mem += child.memory_info()[0]
        return par_cpu, par_mem, n_children, ch_cpu, ch_mem

    def _exceed_resource(self, msg):
        err_file = os.path.join(
            os.path.expanduser("~"), ".sos", "tasks", self.task_id + ".soserr")
        with open(err_file, "a") as err:
            err.write(msg + "\n")
        tf = TaskFile(self.task_id)
        tf.add_outputs()
        tf.status = "aborted"
        env.logger.warning(f"{self.task_id} ``aborted``: {msg}")
        # kill the task
        p = psutil.Process(self.pid)
        p.kill()

    def run(self):
        counter = 0
        start_time = time.time()
        while True:
            try:
                tf = TaskFile(self.task_id)
                if not tf.exists():
                    env.logger.warning(f"Task {self.task_id} ``removed``")
                    # the job should be removed
                    p = psutil.Process(self.pid)
                    p.kill()
                sts = tf.status
                if sts in ("completed", "failed"):
                    break
                if sts == "aborted" or not os.path.isfile(self.pulse_file):
                    env.logger.warning(f"Task {self.task_id} ``aborted``")
                    # the job should be killed
                    p = psutil.Process(self.pid)
                    p.kill()
                    if sts != "aborted":
                        # mark the task file
                        tf.status = "aborted"
                # most of the time we only update
                if counter % self.resource_monitor_interval:
                    os.utime(self.pulse_file, None)
                else:
                    cpu, mem, nch, ch_cpu, ch_mem = self._check()
                    if ("peak_cpu" not in self.sos_dict or
                            self.sos_dict["peak_cpu"] < cpu + ch_cpu):
                        self.sos_dict["peak_cpu"] = cpu + ch_cpu
                    if ("peak_mem" not in self.sos_dict or
                            self.sos_dict["peak_mem"] < mem + ch_mem):
                        self.sos_dict["peak_mem"] = mem + ch_mem

                    with open(self.pulse_file, "a") as pd:
                        pd.write(
                            f"{time.time()}\t{cpu:.2f}\t{mem}\t{nch}\t{ch_cpu}\t{ch_mem}\n"
                        )
                    if self.max_procs is not None and cpu + ch_cpu > self.max_procs:
                        self._exceed_resource(
                            f"Task {self.task_id} exits because of excessive use of procs (used {cpu + ch_cpu}, limit {self.max_procs})"
                        )
                    if self.max_mem is not None and mem + ch_mem > self.max_mem:
                        self._exceed_resource(
                            f"Task {self.task_id} exits because of excessive use of max_mem (used {mem + ch_mem}, limit {self.max_mem})"
                        )
                # walltime can be checked more frequently and does not have to wait for resource option
                elapsed = time.time() - start_time
                if self.max_walltime is not None and elapsed > self.max_walltime:
                    self._exceed_resource(
                        f"Task {self.task_id} exits because of excessive run time (used {format_HHMMSS(int(elapsed))}, limit {format_HHMMSS(self.max_walltime)})"
                    )
                time.sleep(self.monitor_interval)
                counter += 1
            except Exception as e:
                # if the process died, exit the thread
                # the warning message is usually:
                # WARNING: psutil.NoSuchProcess no process found with pid XXXXX
                # env.logger.warning(str(e))
                env.logger.debug(
                    f"Monitor of {self.task_id} failed with message {e}")
                break


class WorkflowMonitor(threading.Thread):

    def __init__(
        self,
        workflow_id,
        monitor_interval,
        resource_monitor_interval,
        max_walltime=None,
        max_mem=None,
        max_procs=None,
        sos_dict={},
    ):
        threading.Thread.__init__(self)
        self.workflow_id = workflow_id
        self.pid = os.getpid()
        self.monitor_interval = monitor_interval
        self.resource_monitor_interval = max(
            resource_monitor_interval // monitor_interval, 1)
        self.daemon = True
        self.max_walltime = max_walltime
        if self.max_walltime is not None:
            self.max_walltime = expand_time(self.max_walltime)
        self.max_mem = max_mem
        self.max_procs = max_procs
        self.pulse_file = os.path.join(
            os.path.expanduser("~"), ".sos", "workflows",
            workflow_id + ".pulse")
        # remove previous status file, which could be readonly if the job is killed
        if os.path.isfile(self.pulse_file):
            if not os.stat(self.pulse_file).st_mode & stat.S_IWUSR:
                os.chmod(self.pulse_file, stat.S_IREAD | stat.S_IWRITE)
            os.remove(self.pulse_file)
        self.sos_dict = sos_dict
        with open(self.pulse_file, "a") as pd:
            pd.write(
                "#time\tproc_cpu\tproc_mem\tchildren\tchildren_cpu\tchildren_mem\n"
            )

    def _check(self):
        current_process = psutil.Process(self.pid)
        par_cpu = current_process.cpu_percent()
        par_mem = current_process.memory_info()[0]
        ch_cpu = 0
        ch_mem = 0
        children = current_process.children(recursive=True)
        n_children = len(children)
        for child in children:
            ch_cpu += child.cpu_percent()
            ch_mem += child.memory_info()[0]
        return par_cpu, par_mem, n_children, ch_cpu, ch_mem

    def _exceed_resource(self, msg):
        err_file = os.path.join(
            os.path.expanduser("~"), ".sos", "workflows",
            self.workflow_id + ".soserr")
        with open(err_file, "a") as err:
            err.write(msg + "\n")
        env.logger.warning(f"{self.workflow_id} ``aborted``: {msg}")
        # kill the workflow
        p = psutil.Process(self.pid)
        p.kill()

    def write(self, msg):
        with open(self.pulse_file, "a") as pd:
            pd.write(f"#{time.time()}\t{msg}\n")

    def run(self):
        counter = 0
        start_time = time.time()
        while True:
            try:
                if not os.path.isfile(self.pulse_file) or not os.stat(
                        self.pulse_file).st_mode & stat.S_IWUSR:
                    env.logger.warning(
                        f"Workflow {self.workflow_id} ``aborted``")
                    # the job should be killed
                    p = psutil.Process(self.pid)
                    p.kill()
                # most of the time we only update
                if counter % self.resource_monitor_interval:
                    os.utime(self.pulse_file, None)
                else:
                    cpu, mem, nch, ch_cpu, ch_mem = self._check()
                    if ("peak_cpu" not in self.sos_dict or
                            self.sos_dict["peak_cpu"] < cpu + ch_cpu):
                        self.sos_dict["peak_cpu"] = cpu + ch_cpu
                    if ("peak_mem" not in self.sos_dict or
                            self.sos_dict["peak_mem"] < mem + ch_mem):
                        self.sos_dict["peak_mem"] = mem + ch_mem

                    with open(self.pulse_file, "a") as pd:
                        pd.write(
                            f"{time.time()}\t{cpu:.2f}\t{mem}\t{nch}\t{ch_cpu}\t{ch_mem}\n"
                        )
                    if self.max_procs is not None and cpu + ch_cpu > self.max_procs:
                        self._exceed_resource(
                            f"Workflow {self.workflow_id} exits because of excessive use of procs (used {cpu + ch_cpu}, limit {self.max_procs})"
                        )
                    if self.max_mem is not None and mem + ch_mem > self.max_mem:
                        self._exceed_resource(
                            f"Workflow {self.workflow_id} exits because of excessive use of max_mem (used {mem + ch_mem}, limit {self.max_mem})"
                        )
                # walltime can be checked more frequently and does not have to wait for resource option
                elapsed = time.time() - start_time
                if self.max_walltime is not None and elapsed > self.max_walltime:
                    self._exceed_resource(
                        f"Workflow {self.workflow_id} exits because of excessive run time (used {format_HHMMSS(int(elapsed))}, limit {format_HHMMSS(self.max_walltime)})"
                    )
                time.sleep(self.monitor_interval)
                counter += 1
            except Exception as e:
                # if the process died, exit the thread
                # the warning message is usually:
                # WARNING: psutil.NoSuchProcess no process found with pid XXXXX
                # env.logger.warning(str(e))
                env.logger.debug(
                    f"Monitor of {self.workflow_id} failed with message {e}")
                break


def summarizeExecution(task_id, pulses, status="Unknown"):
    peak_cpu = 0
    accu_cpu = 0
    peak_mem = 0
    accu_mem = 0
    peak_nch = 0
    start_time = None
    end_time = None
    count = 0
    for line in pulses.splitlines():
        if line.startswith("#"):
            continue
        try:
            t, c, m, nch, cc, cm = line.split()
        except Exception as e:
            env.logger.warning(
                f'Unrecognized resource line "{line.strip()}": {e}')
        if start_time is None:
            start_time = float(t)
            end_time = float(t)
        else:
            end_time = float(t)
        accu_cpu += float(c) + float(cc)
        accu_mem += float(m) + float(cm)
        count += 1
        if float(c) + float(cc) > peak_cpu:
            peak_cpu = float(c) + float(cc)
        if float(m) + float(cm) > peak_mem:
            peak_mem = float(m) + float(cm)
        if int(nch) > peak_nch:
            peak_nch = int(nch)
    try:
        second_elapsed = end_time - start_time
    except Exception:
        second_elapsed = 0
    result = [
        ("status", status),
        ("task", task_id),
        ("nproc", str(peak_nch)),
        ("start", time.strftime("%Y-%m-%d %H:%M:%S",
                                time.localtime(start_time))),
        ("end", time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(end_time))),
        (
            "duration",
            ("" if second_elapsed < 86400 else
             f'{int(second_elapsed/86400)} day{"s" if second_elapsed > 172800 else ""} '
            ) + time.strftime("%H:%M:%S", time.gmtime(second_elapsed)),
        ),
        ("cpu_peak", f"{peak_cpu:.1f}"),
        ("cpu_avg", f"{0 if count == 0 else accu_cpu/count:.1f}"),
        ("mem_peak", f"{peak_mem/1024/1024:.1f}Mb"),
        ("mem_avg", f"{0 if count == 0 else accu_mem/1024/1024/count:.1f}Mb"),
    ]
    return "\n".join(f"{x:20s} {y}" for x, y in result)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.
import copy
import os
import stat
import subprocess
import time

from .eval import cfg_interpolate
from .utils import env, expand_time, format_duration, textMD5


class WorkflowEngine:

    def __init__(self, agent):
        self.agent = agent
        self.config = agent.config
        self.alias = self.config["alias"]

    def remove_arg(self, argv, arg):
        r_idx = [idx for idx, x in enumerate(argv) if x.startswith(arg)]
        if not r_idx:
            return argv
        else:
            r_idx = r_idx[0]
        # find next option
        r_next = [
            idx for idx, x in enumerate(argv[r_idx + 1:]) if x.startswith("-")
        ]
        if r_next:
            argv = argv[:r_idx] + argv[r_idx + 1 + r_next[0]:]
        else:
            argv = argv[:r_idx]
        return argv

    def expand_template(self):
        try:
            self.template_args["filename"] = self.filename
            self.template_args["command"] = self.command
            self.template_args["job_name"] = self.job_name
            self.job_text = (
                cfg_interpolate(self.workflow_template, self.template_args) +
                "\n")
        except Exception as e:
            raise ValueError(
                f"Failed to generate job file for the execution of workflow: {e}"
            )
        try:
            wf_dir = os.path.join(os.path.expanduser("~"), ".sos", "workflows")
            if not os.path.isdir(wf_dir):
                os.makedirs(wf_dir)

            self.job_file = os.path.join(wf_dir, self.job_name + ".sh")

            # do not translate newline under windows because the script will be executed
            # under linux/mac
            with open(self.job_file, "w", newline="") as job:
                job.write(self.job_text)
        except Exception as e:
            raise RuntimeError(
                f"Failed to submit workflow {self.command} with script \n{self.job_text}\n: {e}"
            )
        return True

    def execute_workflow(self, filename, command, **template_args):
        # there is no need to prepare workflow (e.g. copy over)
        from .__main__ import get_run_parser
        from .parser import SoS_Script

        parser = get_run_parser(interactive=False, with_workflow=True)
        args, workflow_args = parser.parse_known_args(command[2:])
        script = SoS_Script(filename=filename)
        workflow = script.workflow(
            args.workflow, use_default=not args.__targets__)

        # we send config files also to remote host, but change localhost
        import yaml

        # process -c configfile
        host_cfg_file = os.path.join(
            os.path.expanduser("~"), ".sos", f"config_{self.alias}.yml")
        with open(host_cfg_file, "w") as cfg:
            remote_cfg = copy.deepcopy(env.sos_dict["CONFIG"])
            remote_cfg["localhost"] = self.alias
            yaml.safe_dump(remote_cfg, cfg)
            cfg.flush()
            # copy the files over
            self.agent.send_job_file(f"~/.sos/config_{self.alias}.yml", dir=".")

        self.local_filename = filename

        self.job_name = workflow.calc_md5(workflow_args)

        if os.path.isfile(filename):
            ret = self.agent.send_to_host([filename])
        elif os.path.isfile(filename + ".sos"):
            ret = self.agent.send_to_host([filename + ".sos"])
        elif os.path.isfile(filename + ".ipynb"):
            ret = self.agent.send_to_host([filename + ".ipynb"])

        self.filename = list(ret.values())[0]
        self.command = self.remove_arg(command, "-r")
        # -c only point to local config file.
        self.command = self.remove_arg(self.command, "-c")
        self.command += ["-c", f"~/.sos/config_{self.alias}.yml"]
        # remove --slave mode because the master cannot reach remote slave
        self.command = self.remove_arg(self.command, "-m")
        self.command += ["-M", self.job_name]
        # replace absolute path with relative one because remote sos might have
        # a different path.
        if os.path.basename(command[0]) == "sos":
            self.command[0] = "sos"
            self.command[2] = self.filename
        elif os.path.basename(command[0]) == "sos-runner":
            self.command[0] = "sos-runner"
            self.command[1] = self.filename
        else:
            raise ValueError(
                f"Failed to generate remote execution command: {self.command}")
        self.command = subprocess.list2cmdline(self.command)
        self.template_args = copy.deepcopy(self.config)
        self.template_args.update(template_args)
        env.log_to_file("WORKFLOW",
                        f"Execute command on remote host: {self.command}")
        return True

    def query_workflows(
        self,
        workflows=None,
        check_all=False,
        verbosity=1,
        html=False,
        numeric_times=False,
        age=None,
        tags=None,
        status=None,
    ):
        try:
            return self.agent.check_output(
                "{} status {} -v {} {} {} {} {} {} {}".format(
                    self.agent.config.get("sos", "sos"),
                    "" if workflows is None else " ".join(workflows),
                    verbosity,
                    "--all workflows" if check_all else "",
                    "--html" if html else "",
                    "--numeric-times" if numeric_times else "",
                    f"--age {age}" if age else "",
                    f'--tags {" ".join(tags)}' if tags else "",
                    f'--status {" ".join(status)}' if status else "",
                ))
        except subprocess.CalledProcessError as e:
            if verbosity >= 3:
                env.logger.warning(
                    f'Failed to query status of workflows on {self.alias}: {"" if e.stderr is None else e.stderr.decode()}'
                )
            return ""

    def kill_workflows(self, workflows, tags=None, all_workflows=False):
        cmd = "{} kill {} {} {}".format(
            self.agent.config.get("sos", "sos"),
            "" if all_workflows else " ".join(workflows),
            f'--tags {" ".join(tags)}' if tags else "",
            "--all workflows" if all_workflows else "",
        )

        try:
            ret = self.agent.check_output(cmd)
            env.logger.debug(f'"{cmd}" executed with response "{ret}"')
        except subprocess.CalledProcessError:
            env.logger.error(
                "Failed to kill all workflows" if all_workflows else
                f'Failed to kill workflows {" ".join(workflows)}')
            return ""
        return ret

    def purge_workflows(self,
                        workflows,
                        purge_all=False,
                        age=None,
                        status=None,
                        tags=None,
                        verbosity=2):
        try:
            return self.agent.check_output(
                "{} purge {} {} {} {} {} -v {}".format(
                    self.agent.config.get("sos", "sos"),
                    " ".join(workflows),
                    "--all" if purge_all else "",
                    f"--age {age}" if age is not None else "",
                    f'--status {" ".join(status)}'
                    if status is not None else "",
                    f'--tags {" ".join(tags)}' if tags is not None else "",
                    verbosity,
                ))
        except subprocess.CalledProcessError:
            env.logger.error(f"Failed to purge workflows {workflows}")
            return ""


class BackgroundProcess_WorkflowEngine(WorkflowEngine):

    def __init__(self, agent):
        super(BackgroundProcess_WorkflowEngine, self).__init__(agent)
        if "workflow_template" in self.config:
            self.workflow_template = self.config["workflow_template"].replace(
                "\r\n", "\n")
        else:
            self.workflow_template = None

    def execute_workflow(self, filename, command, **template_args):
        #
        # calling super execute_workflow would set cleaned versions
        # of self.filename, self.command, and self.template_args
        if not super(BackgroundProcess_WorkflowEngine, self).execute_workflow(
                filename, command, **template_args):
            env.log_to_file(
                "WORKFLOW",
                f'Failed to prepare workflow with command "{command}"')
            return False

        if self.workflow_template:
            if not self._execute_workflow_with_template():
                return False
        else:
            if not self._execute_workflow():
                return False
        return True

    def _execute_workflow(self):
        # if no template, use a default command
        env.log_to_file("WORKDLOW", f'Execute "{self.command}"')
        try:
            self.agent.check_call(self.command, under_workdir=True)
        except Exception as e:
            raise RuntimeError(f"Failed to submit workflow {self.command}: {e}")
        return True

    def _execute_workflow_with_template(self):
        """Submit workflows by interpolating a shell script defined in workflow_template"""
        self.expand_template()

        try:
            # then copy the job file to remote host if necessary
            self.agent.send_job_file(self.job_file, dir="workflows")

            cmd = f"bash ~/.sos/workflows/{os.path.basename(self.job_file)}"
            env.log_to_file(
                "WORKFLOW",
                f'Execute "{self.command}" with script {self.job_text}')
            self.agent.check_call(cmd, under_workdir=True)
        except Exception as e:
            raise RuntimeError(
                f"Failed to submit workflow {self.command} with script \n{self.job_text}\n: {e}"
            )
        finally:
            try:
                os.remove(self.job_file)
            except Exception as e:
                env.logger.debug(
                    f"Failed to remove temporary workflow file: {e}")
        return True


class WorkflowPulse:

    def __init__(self, workflow_id):
        self.id = workflow_id
        self._tags = None
        self._status = None
        self._content = {}
        self._start_time = None
        self._complete_time = None
        self.pulse_file = os.path.join(
            os.path.expanduser("~"), ".sos", "workflows", self.id + ".pulse")

    def exists(self):
        return os.path.isfile(self.pulse_file)

    def mark_killed(self):
        from stat import S_IREAD

        os.chmod(self.pulse_file, S_IREAD)

    @property
    def created(self):
        return ("Created " +
                format_duration(time.time() - self._start_time, True) +
                " ago" if self._start_time else "")

    @property
    def started(self):
        return ("Started " +
                format_duration(time.time() - self._start_time, True) +
                " ago" if self._start_time else "")

    @property
    def duration(self):
        if not self._complete_time:
            return ""
        return "Ran for " + format_duration(
            int(self._complete_time - self._start_time))

    @property
    def tags(self):
        if self._tags is None:
            self.parse_pulse_file()
        return self._tags

    @property
    def status(self):
        if self._status is None:
            self.parse_pulse_file()
        return self._status

    def get_file(self, ext):
        if ext in self._content:
            return self._content[ext]

        res_file = os.path.join(
            os.path.expanduser("~"), ".sos", "workflows", self.id + ext)
        content = ""
        if os.path.isfile(res_file):
            try:
                with open(res_file) as res:
                    content = res.read()
            except:
                pass
        self._content[ext] = content
        return content

    @property
    def script(self):
        return self.get_file(".sos")

    @property
    def stdout(self):
        return self.get_file(".out")

    @property
    def stderr(self):
        return self.get_file(".err")

    @property
    def sh_script(self):
        return self.get_file(".sh")

    def parse_pulse_file(self):
        self._status = "unknown"
        self._tags = ""
        last_active_time = None
        if not os.path.isfile(self.pulse_file):
            self._status = 'missing'
            return
        with open(self.pulse_file) as pulse:
            for line in pulse:
                if line.startswith("#time\t"):
                    continue
                if not line.startswith("#"):
                    try:
                        last_active_time = line.split('\t', 1)[0]
                    except:
                        pass
                    continue
                fields = line.split("\t")
                if len(fields) != 3:
                    env.logger.error(line)
                    continue
                last_active_time = fields[0][1:]
                if fields[1] == "status":
                    self._status = fields[2].strip()
                    if self._status == "running":
                        try:
                            self._start_time = float(fields[0][1:])
                        except:
                            pass
                    elif self._status == "completed":
                        try:
                            self._complete_time = float(fields[0][1:])
                        except:
                            pass
                elif fields[1] == "tags":
                    self._tags = fields[2].strip()
        if not os.stat(self.pulse_file).st_mode & stat.S_IWUSR:
            self._status = 'aborted'
        if last_active_time is None or time.time() - float(
                last_active_time) > 120:
            self._status = 'failed'


def print_workflow_status(
    workflows,
    check_all=False,
    verbosity: int = 1,
    html: bool = False,
    numeric_times=False,
    age=None,
    tags=None,
    status=None,
):
    import glob

    all_workflows: List = []
    if check_all:
        workflows = glob.glob(
            os.path.join(
                os.path.expanduser("~"), ".sos", "workflows", "*.pulse"))
        all_workflows = [
            (os.path.basename(x)[:-6], os.path.getmtime(x)) for x in workflows
        ]
        if not all_workflows:
            return
    else:
        for t in workflows:
            matched_names = glob.glob(
                os.path.join(
                    os.path.expanduser("~"), ".sos", "workflows",
                    f"{t}*.pulse"))
            matched = [(os.path.basename(x)[:-6], os.path.getmtime(x))
                       for x in matched_names]
            if not matched:
                all_workflows.append((t, None))
            else:
                all_workflows.extend(matched)

    if age is not None:
        age = expand_time(age, default_unit="d")
        if age > 0:
            all_workflows = [
                x for x in all_workflows if time.time() - x[1] >= age
            ]
        else:
            all_workflows = [
                x for x in all_workflows if time.time() - x[1] <= -age
            ]

    all_workflows = sorted(
        list(set(all_workflows)), key=lambda x: 0 if x[1] is None else x[1])

    if tags:
        all_workflows = [
            x for x in all_workflows if WorkflowPulse(x[0]).exists() and any(
                y in tags for y in WorkflowPulse(x[0]).tags.split())
        ]

    if not all_workflows:
        return

    workflow_info = [WorkflowPulse(x[0]) for x in all_workflows]
    #
    # automatically remove non-running workflows that are more than 30 days old
    to_be_removed = [
        t for s, (t, d) in zip(workflow_info, all_workflows)
        if d is not None and time.time() - d > 30 * 24 * 60 *
        60 and s != "running"
    ]

    if status:
        workflow_info = [x for x in workflow_info if x.status in status]
    #
    if verbosity == 0:
        print("\n".join([x.status for x in workflow_info]))
    elif verbosity == 1:
        for info in workflow_info:
            print(f"{info.id}\t{info.status}")
    elif verbosity == 2:
        tsize = 20
        for info in workflow_info:
            print(
                f"{info.id}\t{info.tags.ljust(tsize)}\t{info.duration:<14}\t{info.status}"
            )
    elif verbosity == 3:
        tsize = 20
        for info in workflow_info:
            tsize = max(tsize, len(info.tags))
            print(
                f"{info.id}\t{info.tags.ljust(tsize)}\t{info.created:<14}\t{info.started:<14}\t{info.duration:<14}\t{info.status}"
            )
    elif verbosity == 4:
        import pprint

        for info in workflow_info:
            if info.status == "missing":
                print(f"{info.id}\t{info.status}\n")
                continue
            print(f"WORKFLOW:\t{info.id}")
            print(f"status\t{info.status}")
            print(f"{info.created}")
            if info.started:
                print(f"{info.started}")
            if info.duration:
                print(f"{info.duration}")

            print()
            print("TAGS:\n=====")
            print(info.tags)
            print()

            if info.script:
                print("CRIPT:\n=====")
                print(info.script)

            if info.sh_script:
                print("WRAPPER SCRIPT:\n==============")
                print(info.sh_script)

            if info.stdout:
                print("STDOUT:\n=======")
                print(info.stdout)

            if info.stderr:
                print("STDERR:\n=======")
                print(info.stderr)

    # remove jobs that are older than 1 month
    if to_be_removed:
        purge_workflows(to_be_removed, verbosity=0)


def kill_workflow(workflow):
    wp = WorkflowPulse(workflow)
    status = wp.status
    if status == "completed":
        return "completed"
    with open(
            os.path.join(
                os.path.expanduser("~"), ".sos", "workflows",
                workflow + ".soserr"),
            "a",
    ) as err:
        err.write(
            f"Workflow {workflow} killed by sos kill command or workflow engine."
        )

    wp.mark_killed()
    return "aborted"


def kill_workflows(workflows, tags=None):
    import glob

    if not workflows:
        workflows = glob.glob(
            os.path.join(
                os.path.expanduser("~"), ".sos", "workflows", "*.pulse"))
        all_workflows = [os.path.basename(x)[:-6] for x in workflows]
    else:
        all_workflows = []
        for t in workflows:
            matched = glob.glob(
                os.path.join(
                    os.path.expanduser("~"), ".sos", "workflows",
                    f"{t}*.pulse"))
            matched = [os.path.basename(x)[:-6] for x in matched]
            if not matched:
                env.logger.warning(f"{t} does not match any existing workflow")
            else:
                all_workflows.extend(matched)
    if tags:
        all_workflows = [
            x for x in all_workflows if any(
                x in tags for x in WorkflowPulse(x).tags.split())
        ]

    if not all_workflows:
        # env.logger.warning("No workflow to kill")
        return
    all_workflows = sorted(list(set(all_workflows)))
    # at most 20 threads
    for wf in all_workflows:
        ret = kill_workflow(wf)
        print(f"{wf}\t{ret}")


def purge_workflows(workflows,
                    purge_all=None,
                    age=None,
                    status=None,
                    tags=None,
                    verbosity=2):
    import glob

    if workflows:
        all_workflows = []
        for t in workflows:
            matched = glob.glob(
                os.path.join(
                    os.path.expanduser("~"), ".sos", "workflows",
                    f"{t}*.pulse"))
            matched = [
                (os.path.basename(x)[:-6], os.path.getmtime(x)) for x in matched
            ]
            if not matched:
                print(f"{t}\tmissing")
            all_workflows.extend(matched)
    elif purge_all or age or status or tags:
        workflows = glob.glob(
            os.path.join(
                os.path.expanduser("~"), ".sos", "workflows", "*.pulse"))
        all_workflows = [
            (os.path.basename(x)[:-6], os.path.getmtime(x)) for x in workflows
        ]
    else:
        raise ValueError(
            "Please specify either workflows or one or more of --all, --status, --tags--age"
        )
    #
    if age is not None:
        age = expand_time(age, default_unit="d")
        if age > 0:
            all_workflows = [
                x for x in all_workflows if time.time() - x[1] >= age
            ]
        else:
            all_workflows = [
                x for x in all_workflows if time.time() - x[1] <= -age
            ]

    if status:
        all_workflows = [
            x for x in all_workflows if WorkflowPulse(x[0]).status in status
        ]

    if tags:
        all_workflows = [
            x for x in all_workflows if any(
                x in tags for x in WorkflowPulse(x[0]).tags.split())
        ]
    #
    # remoe all workflow files
    all_workflows = set([x[0] for x in all_workflows])
    if all_workflows:
        #
        # find all related files, including those in nested directories
        from collections import defaultdict

        to_be_removed = defaultdict(list)
        for dirname, _, filelist in os.walk(
                os.path.join(os.path.expanduser("~"), ".sos", "workflows")):
            for f in filelist:
                ID = os.path.basename(f).split(".", 1)[0]
                if ID in all_workflows:
                    to_be_removed[ID].append(os.path.join(dirname, f))
        #
        status_cache = {}
        for workflow in all_workflows:
            removed = True
            for f in to_be_removed[workflow]:
                try:
                    if verbosity > 3:
                        if ("WORKFLOW" in env.config["SOS_DEBUG"] or
                                "ALL" in env.config["SOS_DEBUG"]):
                            env.log_to_file("TASK", f"Remove {f}")
                    os.remove(f)
                except Exception as e:
                    removed = False
                    if verbosity > 0:
                        env.logger.warning(
                            f"Failed to purge workflow {workflow[0]}: {e}")
            status_cache.pop(workflow, None)
            if removed and verbosity > 1:
                print(f"{workflow}\tpurged")
    elif verbosity > 1:
        env.logger.debug("No matching workflows to purge")
    if purge_all and age is None and status is None and tags is None:
        matched = glob.glob(
            os.path.join(os.path.expanduser("~"), ".sos", "workflows", "*"))
        count = 0
        for f in matched:
            if os.path.isdir(f):
                import shutil

                try:
                    shutil.rmtree(f)
                    count += 1
                except Exception as e:
                    if verbosity > 0:
                        env.logger.warning(f"Failed to remove {f}: {e}")
            else:
                try:
                    os.remove(f)
                    count += 1
                except Exception as e:
                    if verbosity > 0:
                        env.logger.warning(f"Failed to remove {e}")
        if count > 0 and verbosity > 1:
            env.logger.info(f"{count} other files and directories are removed.")
    return ""
